/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

static FLOAT_T get_f_measure(FLOAT_T precision, FLOAT_T recall, FLOAT_T beta)
{
  return
    ((beta * beta + 1) * precision * recall) /
    (beta * beta * precision + recall);
}

static void initialize_eval(void)
{
  int k1,k2,e;

  for (k1=0; k1<num_class; k1++)
    for (k2=0; k2<num_ans_class; k2++)
      cluster_case_freq_joint[k1][k2] = 0;

  for (e=0; e<num_entry; e++) {
    k1 = cluster_case_pred[e];
    k2 = freq_list[e].ans_id;
    cluster_case_freq_joint[k1][k2] += freq_list[e].freq;
  }
}

static void compute_purity(void)
{
  int k1,k2;
  FLOAT_T sum,max;
  int n, **n_joint;

  /* Aliases for readability */
  n_joint = cluster_case_freq_joint;
  n = num_entry;

  sum = 0.0;
  for (k1=0; k1<num_class; k1++) {
    max = n_joint[k1][0];
    for (k2=1; k2<num_ans_class; k2++) /* Note the initial value */
      if (n_joint[k1][k2] > max) max = n_joint[k1][k2];
    sum += max;
  }

  purity = sum / n;

  sum = 0.0;
  for (k2=0; k2<num_ans_class; k2++) {
    max = n_joint[0][k2];
    for (k1=1; k1<num_class; k1++) /* Note the initial value */
      if (n_joint[k1][k2] > max) max = n_joint[k1][k2];
    sum += max;
  }

  purity_inv = sum / n;

  purity_f = get_f_measure(purity, purity_inv, weight_purity);
}

static void compute_rand_index(void)
{
  int k1,k2;
  FLOAT_T m11,m10,m01,m00,m,m_joint,m_pred,m_ans;
  int n,**n_joint,*n_pred,*n_ans;

  /* Aliases for readability */
  n_pred = cluster_case_freq;
  n_ans  = cluster_case_freq_ans;
  n_joint = cluster_case_freq_joint;
  n = num_entry;

  m_joint = 0.0;
  for (k1=0; k1<num_class; k1++)
    for (k2=0; k2<num_ans_class; k2++)
      m_joint += CHOOSE2(n_joint[k1][k2]);

  m_pred = 0.0;
  for (k1=0; k1<num_class; k1++)
    m_pred += CHOOSE2(n_pred[k1]);

  m_ans = 0.0;
  for (k2=0; k2<num_ans_class; k2++)
    m_ans += CHOOSE2(n_ans[k2]);

  m = CHOOSE2(n);

  m11 = m_joint;
  m10 = m_pred - m11;
  m01 = m_ans - m11;
  m00 = m - (m11 + m10 + m01);

  rand_index = (m11 + m00) / m;

  wallance1 = m11 / (m11 + m10);
  wallance2 = m11 / (m11 + m01);
  wallance_f = get_f_measure(wallance1, wallance2, weight_wallance);

  fowlkes_mallows = SQRT(wallance1 * wallance2);

  adj_rand_index =
    (m_joint - m_pred * m_ans / m) /
    (0.5 * (m_pred + m_ans) - m_pred * m_ans / m);

  jaccard = m11 / (m11 + m10 + m01);

  mirkin = 2 * m * (1.0 - rand_index);
}

static void compute_nmi(void)
{
  int k1,k2;
  FLOAT_T mi,ent_pred,ent_ans;
  FLOAT_T p;
  int n, **n_joint, *n_pred, *n_ans;

  /* Aliases for readability */
  n_pred = cluster_case_freq;
  n_ans  = cluster_case_freq_ans;
  n_joint = cluster_case_freq_joint;
  n = num_entry;

  mi = 0.0;
  for (k1=0; k1<num_class; k1++)
    for (k2=0; k2<num_ans_class; k2++)
      if (n_joint[k1][k2] > 0 && n_pred[k1] > 0 && n_ans[k2] > 0)
        mi +=
          ((FLOAT_T)n_joint[k1][k2] / n) *
          LOG2((FLOAT_T)n_joint[k1][k2] * n / (n_pred[k1] * n_ans[k2]));

  ent_pred = 0.0;
  for (k1=0; k1<num_class; k1++) {
    p = (FLOAT_T)n_pred[k1] / n;
    if (p > 0) ent_pred -= p * LOG2(p);
  }

  ent_ans = 0.0;
  for (k2=0; k2<num_ans_class; k2++) {
    p = (FLOAT_T)n_ans[k2] / n;
    if (p > 0) ent_ans -= p * LOG2(p);
  }

  nmi_score = 2 * mi / (ent_pred + ent_ans);
  vi_score  = ent_pred + ent_ans - 2 * mi;
  nvi_score = 1.0 - nmi_score;
}

void eval_main(void)
{
  FILE *fp_eval;

  alloc_eval();
  initialize_eval();

  compute_purity();
  compute_rand_index();
  compute_nmi();

  fp_eval = fopen(fname_eval,"wb");
  if (fp_eval == NULL) die("Couldn't open %s\n",fname_eval);
  print_eval(fp_eval);
  fclose(fp_eval);

  release_eval();
}
