/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

static void check_gauss(int n, FLOAT_T d)
{
  if (ISNAN(d)) {
    print_param(fp_log);
    DIE_NUMERIC("Computed Gaussian p.d.f. is not a number\n");
  }
  if (ISINF(d)) {
    print_param(fp_log);
    DIE_NUMERIC("Computed Gaussian p.d.f. is not finite\n");
  }
}

/* var should be positive here */
static FLOAT_T lgauss_uv(FLOAT_T x, FLOAT_T u, FLOAT_T var)
{
  return ((-0.5) * (x - u)*(x - u) / var) - 0.5 * LOG(2.0 * M_PI * var);
}

static FLOAT_T lgauss_bv(VECTOR x, VECTOR u, MATRIX sigma)
{
  FLOAT_T S0,S1,rho,u0,u1,r,d;

  S0 = SQRT(sigma[0][0]);  /* stddev */
  S1 = SQRT(sigma[1][1]);  /* stddev */
  rho = sigma[0][1]/(S0 * S1);
  r = (1.0 - rho*rho);

  u0 = (x[0] - u[0])/S0;
  u1 = (x[1] - u[1])/S1;

  d = ((-0.5) * (u0*u0 - 2*rho*u0*u1 + u1*u1) / r)
      - LOG(2.0 * M_PI * S0 * S1 * SQRT(r));

  return d;
}

/* Dimensions are assumed to be checked as valid in advance */
static FLOAT_T lgauss_mv(VECTOR x, VECTOR u, MATRIX sigma)
{
  int n;
  FLOAT_T d,d0,d1,det,q;
  VECTOR x0;
  MATRIX prec;

#ifdef CHECK_DIM
  if ((n = dim(x)) <= 0 || n != dim(u) || n != dim0(sigma) || n != dim1(sigma))
    DIE_DIM;
#else
  n = dim(x);
#endif

  x0 = alloc_vector(n);
  prec = alloc_matrix(n,n);

  diff_vector(x,u,x0);
  det = INVERT(sigma,prec);
  if (err_inv)
    DIE_NUMERIC("Failed to invert covariances\n");

  q = qform(prec,x0);

  d0 = (-0.5) * q;
  d1 = 0.5 * (n * LOG(2 * M_PI) + LOG(det));

  d = d0 - d1;

  release_vector(x0);
  release_matrix(prec);

  return d;
}

/* Dimensions are assumed to be checked as valid in advance */
static FLOAT_T
lgauss_mv_prec(VECTOR x, VECTOR u, MATRIX sigma, MATRIX prec, FLOAT_T det)
{
  int n;
  FLOAT_T d,d0,d1,q;
  VECTOR x0;

#ifdef CHECK_DIM
  if ((n = dim(x)) <= 0 || n != dim(u) || n != dim0(sigma) || n != dim1(sigma))
    DIE_DIM;
#else
  n = dim(x);
#endif

  x0 = alloc_vector(n);
  diff_vector(x,u,x0);

  q = qform(prec,x0);

  d0 = (-0.5) * q;
  d1 = 0.5 * (n * LOG(2 * M_PI) + LOG(det));

  d = d0 - d1;

  release_vector(x0);

  return d;
}

/* Wrapper for Gaussians */
FLOAT_T lgauss(VECTOR x, VECTOR u, MATRIX sigma)
{
  int n;
  FLOAT_T d;

#ifdef CHECK_DIM
  if ((n = dim(x)) <= 0 || n != dim(u) || n != dim0(sigma) || n != dim1(sigma))
    DIE_DIM;
#else
  n = dim(x);
#endif

  if (n == 1)
    d = lgauss_uv(x[0],u[0],sigma[0][0]);
  else if (n == 2)
    d = lgauss_bv(x,u,sigma);
  else
    d = lgauss_mv(x,u,sigma);

  check_gauss(n,d);

  return d;
}

FLOAT_T lgauss_prec(VECTOR x, VECTOR u, MATRIX sigma, MATRIX prec, FLOAT_T det)
{
  int n;
  FLOAT_T d;

#ifdef CHECK_DIM
  if ((n = dim(x)) <= 0 || n != dim(u) || n != dim0(prec) || n != dim1(prec))
    DIE_DIM;
#else
  n = dim(x);
#endif

  if (n == 1)
    d = lgauss_uv(x[0],u[0],sigma[0][0]);
  else if (n == 2)
    d = lgauss_bv(x,u,sigma);
  else
    d = lgauss_mv_prec(x,u,sigma,prec,det);

  check_gauss(n,d);

  return d;
}

/* Wrapper for Gaussians */
FLOAT_T gauss(VECTOR x, VECTOR u, MATRIX sigma)
{
  return EXP(lgauss(x,u,sigma));
}

FLOAT_T gauss_prec(VECTOR x, VECTOR u, MATRIX sigma, MATRIX prec, FLOAT_T det)
{
  return EXP(lgauss_prec(x,u,sigma,prec,det));
}

/* Marginalized Gaussian */
FLOAT_T lgauss_marginal(VECTOR x, VECTOR u, MATRIX sigma, int *obs)
{
  int n,i,m;
  VECTOR x_part,u_part;
  MATRIX sigma_part;
  FLOAT_T d;

#ifdef CHECK_DIM
  if ((n = dim(x)) <= 0 || n != dim(u) || n != dim0(sigma)  || n != dim1(sigma))
    DIE_DIM;
#else
  n = dim(x);
#endif

  m = 0;
  for (i=0; i<n; i++)
    if (obs[i]) m++;

  x_part = alloc_vector(m);
  u_part = alloc_vector(m);
  sigma_part = alloc_matrix(m,m);

  proj_vector(x,obs,x_part);
  proj_vector(u,obs,u_part);
  proj_matrix(sigma,obs,obs,sigma_part);

  d = lgauss(x_part,u_part,sigma_part);

  release_vector(x_part);
  release_vector(u_part);
  release_matrix(sigma_part);

  return d;
}

FLOAT_T gauss_marginal(VECTOR x, VECTOR u, MATRIX sigma, int *obs)
{
  return EXP(lgauss_marginal(x,u,sigma,obs));
}

/* Compute log N(x0|x1), where cond is a 0-1 vector:
 *  - `1' indicates the conditional part (x1)
 *  - `0' indicates the query part (x0)
 */
FLOAT_T lgauss_cond(VECTOR x, VECTOR u, MATRIX sigma, int *cond)
{
  FLOAT_T d;
  int n,i,n0,n1;
  int *query;
  VECTOR x0,x1,u0,u1,uu0;
  MATRIX s00,s01,s10,s11,s11_inv,ss00,z;

#ifdef CHECK_DIM
  if ((n = dim(x)) <= 0 || n != dim(u) || n != dim0(sigma)  || n != dim1(sigma))
    DIE_DIM;
#else
  n = dim(x);
#endif

  query = (int *)xmalloc(sizeof(int)*n);

  n1 = 0;
  for (i=0; i<n; i++) {
    if (cond[i]) {
      query[i] = 0;
      n1++;
    }
    else {
      query[i] = 1;
    }
  }
  n0 = n - n1;

  x0 = alloc_vector(n0);
  x1 = alloc_vector(n1);
  u0 = alloc_vector(n0);
  u1 = alloc_vector(n1);
  s00 = alloc_matrix(n0,n0);
  s01 = alloc_matrix(n0,n1);
  s10 = alloc_matrix(n1,n0);
  s11 = alloc_matrix(n1,n1);
  z = alloc_matrix(n0,n1);
  uu0 = alloc_vector(n0);
  s11_inv = alloc_matrix(n1,n1);
  ss00 = alloc_matrix(n0,n0);

  proj_vector(x,query,x0);
  proj_vector(x,cond,x1);
  proj_vector(u,query,u0);
  proj_vector(u,cond,u1);

  proj_matrix(sigma,query,query,s00);
  proj_matrix(sigma,query,cond,s01);
  proj_matrix(sigma,cond,query,s10);
  proj_matrix(sigma,cond,cond,s11);

  INVERT(s11,s11_inv);
  if (err_inv)
    DIE_NUMERIC("Failed to invert the covariances of the observed part");
  prod_matrix(s01,s11_inv,z);

  diff_vector(x1,u1,NULL);
  prod_matrix_vector(z,x1,uu0);
  sum_vector(uu0,u0,NULL);

  prod_matrix(z,s10,ss00);
  diff_matrix(s00,ss00,NULL);

  d = lgauss(x0,uu0,s00);

  release_vector(x0);
  release_vector(x1);
  release_vector(u0);
  release_vector(u1);
  release_matrix(s00);
  release_matrix(s01);
  release_matrix(s10);
  release_matrix(s11);
  release_matrix(z);
  release_vector(uu0);
  release_matrix(s11_inv);
  release_matrix(ss00);

  xfree(query);

  return d;
}

FLOAT_T gauss_cond(VECTOR x, VECTOR u, MATRIX sigma, int *cond)
{
  return EXP(lgauss_cond(x,u,sigma,cond));
}

static FLOAT_T lgauss_stat_uv(FLOAT_T smpl_size, FLOAT_T smpl_u,
                              FLOAT_T smpl_s, FLOAT_T mean, FLOAT_T var)
{
  FLOAT_T d = 0.0;

  d -= 0.5 * (smpl_s - 2 * mean * smpl_u + smpl_size * mean * mean) / var;
  d -= 0.5 * smpl_size * LOG(2.0 * M_PI * var);

  return d;
}

static FLOAT_T lgauss_stat_bv_prec(FLOAT_T smpl_size, VECTOR smpl_u,
                                   MATRIX smpl_s, VECTOR mu, MATRIX sigma,
                                   MATRIX prec, FLOAT_T det)
{
  FLOAT_T q;

  q =
    (smpl_s[0][0] - 2*smpl_u[0]*mu[0] + smpl_size*mu[0]*mu[0]) * prec[0][0] +
    (smpl_s[1][1] - 2*smpl_u[1]*mu[1] + smpl_size*mu[1]*mu[1]) * prec[1][1];

  return ((-0.5) * q) - smpl_size * (LOG(2.0 * M_PI) + 0.5 * LOG(det));
}

static FLOAT_T lgauss_stat_mv_prec(FLOAT_T smpl_size, VECTOR smpl_u,
                                   MATRIX smpl_s, VECTOR mu, MATRIX sigma,
                                   MATRIX prec, FLOAT_T det)
{
  int n,i,j;
  MATRIX prod,smpl_s1;
  FLOAT_T d,tr;

  n = dim0(smpl_s);
  smpl_s1 = alloc_matrix(n,n);
  prod = alloc_matrix(n,n);

  for (i=0; i<n; i++)
    for (j=i; j<n; j++)
      smpl_s1[i][j] =
        smpl_s[i][j]
        - smpl_u[i] * mu[j]
        - mu[i] * smpl_u[j]
        + smpl_size * mu[i] * mu[j];

  copy_upper_to_lower(smpl_s1);

  prod_matrix(smpl_s1,prec,prod);
  tr = trace(prod);

  release_matrix(smpl_s1);
  release_matrix(prod);

  d = - 0.5 * (tr + smpl_size * (n * LOG(2.0 * M_PI) + LOG(det)));

  return d;
}

FLOAT_T lgauss_stat_prec(FLOAT_T smpl_size, VECTOR smpl_u,
                         MATRIX smpl_s, VECTOR mu, MATRIX sigma,
                         MATRIX prec, FLOAT_T det)
{
  int n;
  FLOAT_T d;

  n = dim0(smpl_s);

  if (n == 1)
    d = lgauss_stat_uv(smpl_size,smpl_u[0],smpl_s[0][0],mu[0],sigma[0][0]);
  else if (n == 2)
    d = lgauss_stat_bv_prec(smpl_size,smpl_u,smpl_s,mu,sigma,prec,det);
  else
    d = lgauss_stat_mv_prec(smpl_size,smpl_u,smpl_s,mu,sigma,prec,det);

  check_gauss(n,d);

  return d;
}

FLOAT_T gauss_stat_prec(FLOAT_T smpl_size, VECTOR smpl_u,
                        MATRIX smpl_s, VECTOR mu, MATRIX sigma,
                        MATRIX prec, FLOAT_T det)
{
  return EXP(lgauss_stat_prec(smpl_size,smpl_u,smpl_s,mu,sigma,prec,det));
}


static FLOAT_T lgauss_stat_bv(FLOAT_T smpl_size, VECTOR smpl_u,
                              MATRIX smpl_s, VECTOR mu, MATRIX sigma)
{
  MATRIX prec;
  FLOAT_T det,d;

  prec = alloc_matrix(2,2);
  det = INVERT(sigma,prec);

  d = lgauss_stat_bv_prec(smpl_size,smpl_u,smpl_s,mu,sigma,prec,det);

  release_matrix(prec);

  return d;
}

static FLOAT_T lgauss_stat_mv(FLOAT_T smpl_size, VECTOR smpl_u,
                              MATRIX smpl_s, VECTOR mu, MATRIX sigma)
{
  int n;
  MATRIX prec;
  FLOAT_T det,d;

  n = dim0(smpl_s);
  prec = alloc_matrix(n,n);

  det = INVERT(sigma,prec);
  d = lgauss_stat_mv_prec(smpl_size,smpl_u,smpl_s,mu,sigma,prec,det);

  release_matrix(prec);

  return d;
}

FLOAT_T lgauss_stat(FLOAT_T smpl_size, VECTOR smpl_u,
                    MATRIX smpl_s, VECTOR mu, MATRIX sigma)
{
  int n;
  FLOAT_T d;

#ifdef CHECK_DIM
  if ((n = dim0(smpl_s)) <= 0 || n != dim1(smpl_s) ||
      n != dim0(sigma) || n != dim1(sigma))
    DIE_DIM;
#else
  n = dim0(smpl_s);
#endif

  if (n == 1)
    d = lgauss_stat_uv(smpl_size,smpl_u[0],smpl_s[0][0],mu[0],sigma[0][0]);
  else if (n == 2)
    d = lgauss_stat_bv(smpl_size,smpl_u,smpl_s,mu,sigma);
  else
    d = lgauss_stat_mv(smpl_size,smpl_u,smpl_s,mu,sigma);

  check_gauss(n,d);

  return d;
}

FLOAT_T gauss_stat(FLOAT_T smpl_size, VECTOR smpl_u,
                   MATRIX smpl_s, VECTOR mu, MATRIX sigma)
{
  return EXP(lgauss_stat(smpl_size,smpl_u,smpl_s,mu,sigma));
}

/* Stand-alone compilation:
 *   > gcc -g -std=c99 -DMAIN_GAUSS \
 *       -o gauss gauss.c matrix.c allocate_matrix.c misc.c \
 *          print_en.c read_freq.c init.c hash.c allocate.c -lm
 */

#ifdef MAIN_GAUSS

#include "nbc_var.h"

FLOAT_T frand(void)
{
  return (FLOAT_T)random_float();
}

int main(void)
{
  VECTOR u, v;
  MATRIX sigma;
  int n;
  FLOAT_T g,x,y,z;

  initialize_buffer();

  initialize_matrix_pool();
  initialize_vector_pool();

  n = 3;
  u = alloc_vector(n);
  sigma = alloc_matrix(n,n);

  u[0] = 0.8;
  u[1] = 3.1;
  u[2] = 0.1;

  sigma[0][0] =  0.3;
  sigma[0][1] = -0.6;
  sigma[0][2] =  0.1;
  sigma[1][0] = -0.6;
  sigma[1][1] =  1.5;
  sigma[1][2] = -0.1;
  sigma[2][0] =  0.1;
  sigma[2][1] = -0.1;
  sigma[2][2] =  4.2;

  v = alloc_vector(n);

  for (x = 2.0; x <= 2.0; x += 0.5) {
    for (y = 1.0; y <= 1.0; y += 0.5) {
      for (z = -1.5; z <= -1.5; z += 0.5) {
        /*
        for (x = -5.0; x <= 5.0; x += 0.5) {
        for (y = -5.0; y <= 5.0; y += 0.5) {
          for (z = -5.0; z <= 5.0; z += 0.5) {
        */
        v[0] = x;
        v[1] = y;
        v[2] = z;
        g = gauss(v,u,sigma);
        fprintf(stderr,conv_msg("%.Ne %.Ne %.Ne %.Ne\n"),x,y,z,g);
      }
    }
  }

  release_vector(v);
  release_vector(u);
  release_matrix(sigma);

  release_matrix_pool();
  release_vector_pool();
}

#endif

