/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

static void check_bic(FLOAT_T d)
{
  if (ISNAN(d)) {
    print_param(fp_log);
    DIE_NUMERIC("Computed " IC_STR_BIC " is not a number\n");
  }
  if (ISINF(d)) {
    print_param(fp_log);
    DIE_NUMERIC("Computed " IC_STR_BIC " is not finite\n");
  }
}

static void check_cs(FLOAT_T d)
{
  if (ISNAN(d)) {
    print_param(fp_log);
    DIE_NUMERIC("Computed " IC_STR_AUTOCLASS " is not a number\n");
  }
  if (ISINF(d)) {
    print_param(fp_log);
    DIE_NUMERIC("Computed " IC_STR_AUTOCLASS " is not finite\n");
  }
}

static FLOAT_T compute_complete_marginal_loglike_gauss(void)
{
  int k,u,a0,a1,n;
  FLOAT_T ll,l;
  FLOAT_T smooth_u0,smooth_u1,det0,det1,r,tau0,alpha_w0;
  FLOAT_T *e_u,**e_s,e_c_k;
  MATRIX h_s,sigma0;

  ll = 0.0;

  for (k=0; k<num_class; k++) {
    tau0 = get_tau(k);

    e_c_k = expect_c[k];

    r = (tau0 * e_c_k) / (tau0 + e_c_k);

    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];

      alpha_w0 = get_alpha_w(k,u,n);

      e_u = expect_u[k][u];
      e_s = expect_s[k][u];
      h_s = hparam_s[k][u];

      sigma0 = alloc_matrix(n,n);

      for (a0=0; a0<n; a0++) {
        smooth_u0 = hparam_u[k][u][a0];
        for (a1=a0; a1<n; a1++) {
          smooth_u1 = hparam_u[k][u][a1];

          sigma0[a0][a1] = h_s[a0][a1];
          sigma0[a0][a1] += e_s[a0][a1] - e_u[a0] * e_u[a1] / e_c_k;
          sigma0[a0][a1] +=
            r * (smooth_u0 - e_u[a0]/e_c_k) * (smooth_u1 - e_u[a1]/e_c_k);
        }
      }
      copy_upper_to_lower(sigma0);

      det0 = DETERMINANT(h_s);
      det1 = DETERMINANT(sigma0);

      l = 0.0;
      l -= 0.5 * n * e_c_k * LOG(2.0 * M_PI);
      l += 0.5 * n * (LOG(tau0) - LOG(tau0 + e_c_k));
      l += lwishart_coeff(n,alpha_w0);
      l -= lwishart_coeff(n,(alpha_w0 + e_c_k));
      l += 0.5 * alpha_w0 * LOG(det0);
      l -= 0.5 * (alpha_w0 + e_c_k) * LOG(det1);

      ll += l;

      release_matrix(sigma0);
    }
  }

  return ll;
}

/*
 * Compute the marginal likelihood with complete data
 * (Cooper and Herskovits, 1992)
 *
 * -- expect_c[num_class],
 *    expect_w1[num_class][num_word]
 *    expect_w2[num_class][num_word]
 *      are assumed to be already computed
 */

static FLOAT_T compute_complete_marginal_loglike1(void)
{
  int k,v,i;
  FLOAT_T l;
  FLOAT_T a_sum0,a_sum1;

  a_sum0 = 0.0;
  a_sum1 = 0.0;
  for (k=0; k<num_class; k++) {
    a_sum0 += hparam_c[k];
    a_sum1 += expect_c[k] + hparam_c[k];
  }

  l = LGAMMA(a_sum0) - LGAMMA(a_sum1);
  for (k=0; k<num_class; k++) {
    l += LGAMMA(expect_c[k] + hparam_c[k]) - LGAMMA(hparam_c[k]);
  }

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
        a_sum0 = 0.0;
        a_sum1 = 0.0;
        for (i=0; i<num_word[v]; i++) {
          a_sum0 += hparam_w[k][v][i];
          a_sum1 += expect_w[k][v][i] + hparam_w[k][v][i];
        }
        l += LGAMMA(a_sum0) - LGAMMA(a_sum1);
        for (i=0; i<num_word[v]; i++) {
          l +=
            LGAMMA(expect_w[k][v][i] + hparam_w[k][v][i])
            - LGAMMA(hparam_w[k][v][i]);
        }
      }
    }
  }

  if (mode_gauss)
    l += compute_complete_marginal_loglike_gauss();

  return l;
}

static FLOAT_T compute_complete_marginal_loglike2(void)
{
  int k,v,i;
  FLOAT_T a_c,l;
  FLOAT_T *a_w;
  FLOAT_T t1,t2,t3,t4;

  l = 0.0;
  a_c = smooth_c + 1.0;

  t1 =
    LGAMMA(num_class * a_c)
    - LGAMMA(total_freq + num_class * a_c)
    - num_class * LGAMMA(a_c);

  for (k=0; k<num_class; k++)
    t1 += LGAMMA(expect_c[k] + a_c);

  l += t1;

  /*
   *  In the case of tuple_size_d == 2:
   *    t2 = num_class
   *           * (LGAMMA(num_word1 * a_w1) + LGAMMA(num_word2 * a_w2)
   *              - (num_word1 * LGAMMA(a_w1) + num_word2 * LGAMMA(a_w2)));
   */
  if (mode_nominal) {
    t2 = 0.0;
    t3 = 0.0;
    t4 = 0.0;
    a_w = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*tuple_size_d);
    for (v=0; v<tuple_size_d; v++)
      a_w[v] = smooth_w[v] + 1.0;

    for (v=0; v<tuple_size_d; v++)
      t2 += LGAMMA(num_word[v] * a_w[v]);

    for (v=0; v<tuple_size_d; v++)
      t2 -= num_word[v] * LGAMMA(a_w[v]);

    t2 *= num_class;

    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        t3 -= LGAMMA(expect_c[k] + num_word[v] * a_w[v]);

    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          t4 += LGAMMA(expect_w[k][v][i] + a_w[v]);

    xfree(a_w);

    l += (t2 + t3 + t4);
  }

  if (mode_gauss)
    l += compute_complete_marginal_loglike_gauss();

  return l;
}

static FLOAT_T compute_complete_marginal_loglike(void)
{
  FLOAT_T l;

  l =
    (mode_read_hparam) ?
    compute_complete_marginal_loglike1() :
    compute_complete_marginal_loglike2();

  return l;
}

/* Compute the likelihood with the pseudo complete data */
static FLOAT_T compute_complete_loglike(void)
{
  int k,v,i,u,n;
  FLOAT_T ll,det;
  MATRIX p_s,p_p,e_s;
  VECTOR p_u,e_u;

  ll = 0.0;
  for (k=0; k<num_class; k++) {
    ll += expect_c[k] * LOG(param_c[k]);
  }

  if (mode_nominal) {
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          ll += expect_w[k][v][i] * LOG(param_w[k][v][i]);
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];

        p_u = param_u[k][u];
        p_s = param_s[k][u];
        p_p = param_p[k][u];
        det = param_sdet[k][u];

        e_u = expect_u[k][u];
        e_s = expect_s[k][u];

        ll += lgauss_stat_prec(expect_c[k],e_u,e_s,p_u,p_s,p_p,det);
      }
    }
  }

  return ll;
}

FLOAT_T compute_bic(void)
{
  FLOAT_T ll,l;
  int num_param;
  int v;
  int u,n;

  compute_complete_prob();
  compute_incomplete_prob();
  ll = compute_loglike();

  num_param = (num_class - 1);

  if (mode_nominal) {
    for (v=0; v<tuple_size_d; v++)
      num_param += num_class * (num_word[v] - 1);
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];
      num_param += num_class * (n + n * (n - 1) / 2);  /* means and covars */
    }
  }

  l = ll - 0.5 * num_param * LOG(total_freq);
  check_bic(l);

  return l;
}

FLOAT_T compute_cs(void)
{
  FLOAT_T ll;
  FLOAT_T l0,l1,l2;
  FLOAT_T l;

  compute_complete_prob();
  compute_incomplete_prob();
  compute_conditional_prob();
  compute_expect();
  ll = compute_loglike();

  l0 = compute_complete_marginal_loglike();
  l1 = compute_complete_loglike();
  l2 = ll;

#if 0
  print_msg("l0 = %.Nf\n",l0);
  print_msg("l1 = %.Nf\n",l1);
  print_msg("l2 = %.Nf\n",l2);
#endif

  l = (l0 - l1 + l2);
  check_cs(l);

  return l;
}
