/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

static FLOAT_T determinant1(MATRIX a)
{
  return a[0][0];
}

static FLOAT_T determinant2(MATRIX a)
{
  return a[0][0]*a[1][1] - a[0][1]*a[1][0];
}

static FLOAT_T determinant3(MATRIX a)
{
  return
    a[0][0]*a[1][1]*a[2][2]
    + a[1][0]*a[2][1]*a[0][2]
    + a[2][0]*a[0][1]*a[1][2]
    - a[0][0]*a[2][1]*a[1][2]
    - a[2][0]*a[1][1]*a[0][2]
    - a[1][0]*a[0][1]*a[2][2];
}

static FLOAT_T determinant4(MATRIX a)
{
  int n;
  int *p;
  MATRIX u,l;
  FLOAT_T d;

  err_det = 0;
  n = dim0(a);

  p = (int *)xmalloc(sizeof(int)*n);

  u = alloc_matrix(n,n);
  l = alloc_matrix(n,n);

  d = lu(a,l,u,p);

  if (err_lu) {
    err_det = 1;
    release_matrix(u);
    release_matrix(l);
    return 0.0;
  }

  release_matrix(u);
  release_matrix(l);
  xfree(p);

  return d;
}

/* Using Cholesky decomposition */
static FLOAT_T determinant4_c(MATRIX a)
{
  int n;
  MATRIX l;
  FLOAT_T det;

  err_det = 0;
  n = dim0(a);

  l = alloc_matrix(n,n);

  det = cholesky(a,l);
  if (err_cho) {
    err_det = 1;
    release_matrix(l);
    return 0.0;
  }

  release_matrix(l);
  return det;
}

FLOAT_T determinant(MATRIX a)
{
  int n;
  FLOAT_T det;

#ifdef CHECK_DIM
  if ((n = dim0(a)) <= 0 || n != dim1(a)) DIE_DIM;
#else
  n = dim0(a);
#endif

  if (n == 1)
    det = determinant1(a);
  else if (n == 2)
    det = determinant2(a);
  else if (n == 3)
    det = determinant3(a);
  else
    det = determinant4(a);

  return det;
}

FLOAT_T determinant_c(MATRIX a)
{
  int n;
  FLOAT_T det;

#ifdef CHECK_DIM
  if ((n = dim0(a)) <= 0 || n != dim1(a)) DIE_DIM;
#else
  n = dim0(a);
#endif

  if (n == 1)
    det = determinant1(a);
  else if (n == 2)
    det = determinant2(a);
  else if (n == 3)
    det = determinant3(a);
  else
    det = determinant4_c(a);

  return det;
}

FLOAT_T trace(MATRIX a)
{
  int n,i;
  FLOAT_T t;

#ifdef CHECK_DIM
  if ((n = dim0(a)) <= 0 || n != dim1(a)) DIE_DIM;
#else
  n = dim0(a);
#endif

  t = 0.0;
  for (i=0; i<n; i++)
    t += a[i][i];

  return t;
}

void set_diag(MATRIX a, FLOAT_T d)
{
  int n,i;

#ifdef CHECK_DIM
  if ((n = dim0(a)) <= 0 || n != dim1(a)) DIE_DIM;
#else
  n = dim0(a);
#endif

  for (i=0; i<n; i++)
    a[i][i] = d;
}

void set_diag_matrix(MATRIX a, FLOAT_T d)
{
  int n;

#ifdef CHECK_DIM
  if ((n = dim0(a)) <= 0 || n != dim1(a)) DIE_DIM;
#else
  n = dim0(a);
#endif

  clear_matrix(a);
  set_diag(a,d);
}

FLOAT_T qform(MATRIX a, VECTOR x)
{
  int n;
  int i,j;
  FLOAT_T q0,q1;

#ifdef CHECK_DIM
  if ((n = dim(x)) <= 0 || n != dim0(a) || n != dim1(a)) DIE_DIM;
#else
  n = dim(x);
#endif

  q0 = 0;
  for (i = 0; i < n; i++)
    q0 += a[i][i] * x[i] * x[i];

  q1 = 0;
  for (i = 0; i < (n-1); i++)
    for (j = (i+1); j<n; j++)
      q1 += a[i][j] * x[i] * x[j];

  return (q0 + 2 * q1);
}

/* [Note]
 * The part for partial pivoting and the use of the 'p' array (which
 * corresponds to the pivot matrix 'p' returned by lu() in MATLAB) mimic lu.c,
 * described as an exemplar source code for an algorithm book by H. Okumura
 * (ISBN4-87408-414-1).
 */
FLOAT_T lu(MATRIX a, MATRIX l, MATRIX u, int *p)
{
  int n;
  int i,j,k,i0,j0,k0,jmax;
  FLOAT_T det,s=0.0,t,t0,t1;
  FLOAT_T *max_in_row;
  FLOAT_T m[LU_SCALE_LEN];

  err_lu = 0;

#ifdef CHECK_DIM
  if ((n = dim0(a)) <= 0 || n != dim1(a) || n != dim0(l) || n != dim1(l) ||
      n != dim0(u) || n != dim1(u))
    DIE_DIM;
#else
  n = dim0(a);
#endif

  if (n > LU_SCALE_LEN)
    max_in_row = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*n);
  else
    max_in_row = m;

  clear_matrix(l);
  clear_matrix(u);

  for (i=0; i<n; i++) {
    p[i] = i;
    s = 0.0;
    for (j=0; j<n; j++) {
      if ((t = FABS(a[i][j])) > s) s = t;
    }
    if (s < TINY_FLOAT * TINY_FLOAT) {
      /* the following is not thread-safe, but the only remaining task is
       * to halt the program
       */
      err_lu = 1;
      return 0.0;
    }
    max_in_row[i] = s;
  }

  det = 1;

  for (i=0; i<n; i++) {
    jmax = i;
    for (j=i; j<n; j++) {
      t = FABS(a[p[j]][i]) / max_in_row[p[j]];
      if (t > s) {
        s = t;
        jmax = j;
      }
    }
    if (jmax != i) {
      i0 = p[jmax];
      p[jmax] = p[i];
      p[i] = i0;
      det *= -1;
    }
    else i0 = p[i];

    l[i0][i] = 1;

    t = 0;
    for (k=0; k<i; k++) {
      k0 = p[k];
      t += l[i0][k] * u[k0][i];
    }

    u[i0][i] = (a[i0][i] - t) / l[i0][i];
    if (FABS(u[i0][i]) < TINY_FLOAT * TINY_FLOAT) {
      err_lu = 1;
      return 0.0;
    }

    for (j=(i + 1); j<n; j++) {
      j0 = p[j];
      t0 = 0;
      t1 = 0;
      for (k=0; k<i; k++) {
        k0 = p[k];
        t0 += l[i0][k] * u[k0][j];
        t1 += l[j0][k] * u[k0][i];
      }
      u[i0][j] = (a[i0][j] - t0) / l[i0][i];
      l[j0][i] = (a[j0][i] - t1) / u[i0][i];
    }

    det *= u[i0][i];
  }

  if (n > LU_SCALE_LEN) xfree(max_in_row);

  return det;
}

FLOAT_T cholesky(MATRIX a, MATRIX l)
{
  int n,i,j,k;
  FLOAT_T t,d;

  err_cho = 0;

#ifdef CHECK_DIM
  if ((n = dim0(a)) <= 0 || n != dim1(a) || n != dim0(l) || n != dim1(l))
    DIE_DIM;
#else
  n = dim0(a);
#endif

  d = 1.0;
  for (j=0; j<n; j++) { /* Compute column by column */
    t = a[j][j];
    for (k=0; k<j; k++)
      t -= l[j][k] * l[j][k];

    if (t < 0) {
      err_cho = 1;
      return 0.0;
    }

    d *= (l[j][j] = SQRT(t));
    for (i=(j+1); i<n; i++) {
      t = a[i][j];
      for (k=0; k<j; k++)
        t -= l[i][k] * l[j][k];
      l[i][j] = t / l[j][j];
    }
  }

  return d * d;
}

void invert_l(MATRIX l, MATRIX l_inv)
{
  int n;
  int i,j,k;
  FLOAT_T t;

#ifdef CHECK_DIM
  if ((n = dim0(l)) <= 0 || n != dim1(l) ||
      n != dim0(l_inv) || n != dim1(l_inv))
    DIE_DIM;
#else
  n = dim0(l);
#endif

  clear_matrix(l_inv);

  for (i=0; i<n; i++) {
    l_inv[i][i] = 1.0 / l[i][i];
    for (j=(i-1); j>=0; j--) {
      t = 0;
      for (k=(j + 1); k<=i; k++)
        t += l_inv[i][k] * l[k][j];
      l_inv[i][j] = -t / l[j][j];
    }
  }
}

void invert_l_p(MATRIX l, MATRIX l_inv, int *p)
{
  int n;
  int i,j,k,i0,j0,k0;
  FLOAT_T t;

#ifdef CHECK_DIM
  if ((n = dim0(l)) <= 0 || n != dim1(l) ||
      n != dim0(l_inv) || n != dim1(l_inv))
    DIE_DIM;
#else
  n = dim0(l);
#endif

  clear_matrix(l_inv);

  for (i=0; i<n; i++) {
    i0 = p[i];
    l_inv[i][i] = 1.0 / l[i0][i];
    for (j=(i-1); j>=0; j--) {
      t = 0;
      for (k=(j + 1); k<=i; k++) {
        k0 = p[k];
        t += l_inv[i][k] * l[k0][j];
      }
      j0 = p[j];
      l_inv[i][j] = -t / l[j0][j];
    }
  }
}

void invert_u(MATRIX u, MATRIX u_inv)
{
  int n;
  int i,j,k;
  FLOAT_T t;

#ifdef CHECK_DIM
  if ((n = dim0(u)) <= 0 || n != dim1(u) ||
      n != dim0(u_inv) || n != dim1(u_inv))
    DIE_DIM;
#else
  n = dim0(u);
#endif

  clear_matrix(u_inv);

  for (j=0; j<n; j++) {
    u_inv[j][j] = 1.0 / u[j][j];
    for (i=(j - 1); i>=0; i--) {
      t = 0;
      for (k=(i + 1); k<=j; k++)
        t += u[i][k] * u_inv[k][j];
      u_inv[i][j] = -t / u[i][i];
    }
  }
}

void invert_u_p(MATRIX u, MATRIX u_inv, int *p)
{
  int n;
  int i,j,k,i0,j0;
  FLOAT_T t;

#ifdef CHECK_DIM
  if ((n = dim0(u)) <= 0 || n != dim1(u) ||
      n != dim0(u_inv) || n != dim1(u_inv))
    DIE_DIM;
#else
  n = dim0(u);
#endif

  clear_matrix(u_inv);

  for (j=0; j<n; j++) {
    j0 = p[j];
    u_inv[j][j] = 1.0 / u[j0][j];
    for (i=(j - 1); i>=0; i--) {
      i0 = p[i];
      t = 0;
      for (k=(i + 1); k<=j; k++) {
        t += u[i0][k] * u_inv[k][j];
      }
      u_inv[i][j] = -t / u[i0][i];
    }
  }
}

static FLOAT_T invert1(MATRIX a, MATRIX a_inv)
{
  a_inv[0][0] = 1.0 / a[0][0];

  return a[0][0];
}

static FLOAT_T invert2(MATRIX a, MATRIX a_inv)
{
  FLOAT_T det = determinant2(a);

  a_inv[0][0] =  a[1][1] / det;
  a_inv[0][1] = -a[0][1] / det;
  a_inv[1][0] = -a[1][0] / det;
  a_inv[1][1] =  a[0][0] / det;

  return det;
}

static FLOAT_T invert3(MATRIX a, MATRIX a_inv)
{
  FLOAT_T det = determinant3(a);

  a_inv[0][0] = (a[1][1]*a[2][2] - a[1][2]*a[2][1]) / det;
  a_inv[0][1] = (a[0][2]*a[2][1] - a[0][1]*a[2][2]) / det;
  a_inv[0][2] = (a[0][1]*a[1][2] - a[0][2]*a[1][1]) / det;

  a_inv[1][0] = (a[1][2]*a[2][0] - a[1][0]*a[2][2]) / det;
  a_inv[1][1] = (a[0][0]*a[2][2] - a[0][2]*a[2][0]) / det;
  a_inv[1][2] = (a[0][2]*a[1][0] - a[0][0]*a[1][2]) / det;

  a_inv[2][0] = (a[1][0]*a[2][1] - a[1][1]*a[2][0]) / det;
  a_inv[2][1] = (a[0][1]*a[2][0] - a[0][0]*a[2][1]) / det;
  a_inv[2][2] = (a[0][0]*a[1][1] - a[0][1]*a[1][0]) / det;

  return det;
}

static FLOAT_T invert4(MATRIX a, MATRIX a_inv)
{
  int n;
  int *p,*p_inv;
  MATRIX u,l,u_inv,l_inv,x;
  int i,j,j0;
  FLOAT_T det;

  err_inv = 0;
  n = dim0(a); /* already validated */

  p = (int *)xmalloc(sizeof(int)*n);
  p_inv = (int *)xmalloc(sizeof(int)*n);

  u = alloc_matrix(n,n);
  l = alloc_matrix(n,n);

  det = lu(a,l,u,p);
  if (err_lu) {
    err_inv = 1;
    release_matrix(u);
    release_matrix(l);
    return 0.0;
  }

  u_inv = alloc_matrix(n,n);
  l_inv = alloc_matrix(n,n);

  invert_l_p(l,l_inv,p);
  invert_u_p(u,u_inv,p);

  x = alloc_matrix(n,n);

  prod_matrix(u_inv,l_inv,x);

  release_matrix(u);
  release_matrix(l);
  release_matrix(u_inv);
  release_matrix(l_inv);

  for (i=0; i<n; i++) p_inv[p[i]] = i;

  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) {
      j0 = p_inv[j];
      a_inv[i][j] = x[i][j0];
    }
  }

  release_matrix(x);
  xfree(p);
  xfree(p_inv);

  return det;
}

static FLOAT_T invert4_c(MATRIX a, MATRIX a_inv)
{
  int n;
  MATRIX l,l_inv;
  FLOAT_T det;

  err_inv = 0;
  n = dim0(a);

  l = alloc_matrix(n,n);

  det = cholesky(a,l);
  if (err_cho) {
    err_inv = 1;
    release_matrix(l);
    return 0.0;
  }

  l_inv = alloc_matrix(n,n);

  invert_l(l,l_inv);
  transpose(l_inv,l);  /* l is reused as the transpose of the inverse of l */
  prod_matrix(l,l_inv,a_inv);

  release_matrix(l);
  release_matrix(l_inv);

  return det;
}

FLOAT_T invert(MATRIX a, MATRIX a_inv)
{
  int n;
  FLOAT_T det;

#ifdef CHECK_DIM
  if ((n = dim0(a)) <= 0 || n != dim1(a) ||
      n != dim0(a_inv)   || n != dim1(a_inv))
    DIE_DIM;
#else
  n = dim0(a);
#endif

  if (n == 1)
    det = invert1(a,a_inv);
  else if (n == 2)
    det = invert2(a,a_inv);
  else if (n == 3)
    det = invert3(a,a_inv);
  else
    det = invert4(a,a_inv);

  return det;
}

FLOAT_T invert_c(MATRIX a, MATRIX a_inv)
{
  int n;
  FLOAT_T det;

#ifdef CHECK_DIM
  if ((n = dim0(a)) <= 0 || n != dim1(a)    ||
      n != dim0(a_inv)   || n != dim1(a_inv))
    DIE_DIM;
#else
  n = dim0(a);
#endif

  if (n == 1)
    det = invert1(a,a_inv);
  else if (n == 2)
    det = invert2(a,a_inv);
  else if (n == 3)
    det = invert3(a,a_inv);
  else
    det = invert4_c(a,a_inv);

  return det;
}

void transpose(MATRIX x, MATRIX y)
{
  int n0=-1,n1=-1,i,j;

#ifdef CHECK_DIM
  if ((n0 = dim0(x)) <= 0 || (n1 = dim1(x)) <= 0 ||
      n1 != dim0(y)       || n0 != dim1(y))
    DIE_DIM;
#else
  n0 = dim0(x);
  n1 = dim1(x);
#endif

  for (i=0; i<n0; i++)
    for (j=0; j<n1; j++)
      y[j][i] = x[i][j];
}

/* [Note] We switch the behavior by the nullness of z */
void sum_matrix(MATRIX x, MATRIX y, MATRIX z)
{
  int n0=-1,n1=-1,i,j;

  if (z == NULL) {
#ifdef CHECK_DIM
    if ((n0 = dim0(x)) <= 0 || (n1 = dim1(x)) <= 0 ||
        n0 != dim0(y)       || n1 != dim1(y))
      DIE_DIM;
#else
    n0 = dim0(x);
    n1 = dim1(x);
#endif

    for (i=0; i<n0; i++)
      for (j=0; j<n1; j++)
        x[i][j] += y[i][j];  /* x will be overwritten */
  }
  else {
#ifdef CHECK_DIM
    if ((n0 = dim0(x)) <= 0 || (n1 = dim1(x)) <= 0 ||
        n0 != dim0(y) || n1 != dim1(y) || n0 != dim0(z) || n1 != dim1(z))
      DIE_DIM;
#else
    n0 = dim0(x);
    n1 = dim1(x);
#endif

    for (i=0; i<n0; i++)
      for (j=0; j<n1; j++)
        z[i][j] = x[i][j] + y[i][j];
  }
}

/* [Note] We switch the behavior by the nullness of z */
void diff_matrix(MATRIX x, MATRIX y, MATRIX z)
{
  int n0=-1,n1=-1,i,j;

  if (z == NULL) {
#ifdef CHECK_DIM
    if ((n0 = dim0(x)) <= 0 || (n1 = dim1(x)) <= 0 ||
        n0 != dim0(y) || n1 != dim1(y))
      DIE_DIM;
#else
    n0 = dim0(x);
    n1 = dim1(x);
#endif

    for (i=0; i<n0; i++)
      for (j=0; j<n1; j++)
        x[i][j] -= y[i][j];  /* x will be overwritten */
  }
  else {
#ifdef CHECK_DIM
    if ((n0 = dim0(x)) <= 0 || (n1 = dim1(x)) <= 0 ||
        n0 != dim0(y) || n1 != dim1(y) || n0 != dim0(z) || n1 != dim1(z))
      DIE_DIM;
#else
    n0 = dim0(x);
    n1 = dim1(x);
#endif

    for (i=0; i<n0; i++)
      for (j=0; j<n1; j++)
        z[i][j] = x[i][j] - y[i][j];
  }
}

void prod_matrix(MATRIX a, MATRIX b, MATRIX x)
{
  int n0=-1,n1=-1,n2=-1;
  int i,j,k;

#ifdef CHECK_DIM
  if ((n0 = dim0(a)) <= 0 || (n1 = dim1(a)) <= 0 ||
      n1 != dim0(b) || (n2 = dim1(b)) <= 0 || n0 != dim0(x) || n2 != dim1(x))
    DIE_DIM;
#else
  n0 = dim0(a);
  n1 = dim1(a);
  n2 = dim1(b);
#endif

  clear_matrix(x);

  for (i=0; i<n0; i++)
    for (k=0; k<n1; k++)
      for (j=0; j<n2; j++)
        x[i][j] += a[i][k] * b[k][j];
}

void prod_matrix_vector(MATRIX m, VECTOR v0, VECTOR v1)
{
  int n0=-1,n1=-1,i,j;

#ifdef CHECK_DIM
  if ((n0 = dim0(m)) <= 0 || (n1 = dim1(m)) <= 0 ||
      n1 != dim(v0) || n0 != dim(v1))
    DIE_DIM;
#else
  n0 = dim0(m);
  n1 = dim1(m);
#endif

  clear_vector(v1);

  for (i=0; i<n0; i++)
    for (j=0; j<n1; j++)
      v1[i] += m[i][j] * v0[j];
}

void sum_vector(VECTOR x, VECTOR y, VECTOR z)
{
  int n,i;

  if (z == NULL) {
#ifdef CHECK_DIM
    if ((n = dim(x)) <= 0 || n != dim(y)) DIE_DIM;
#else
    n = dim(x);
#endif

    for (i=0; i<n; i++) x[i] += y[i];
  }
  else {
#ifdef CHECK_DIM
    if ((n = dim(x)) <= 0 || n != dim(y) || n != dim(z)) DIE_DIM;
#else
    n = dim(x);
#endif

    for (i=0; i<n; i++) z[i] = x[i] + y[i];
  }
}

void diff_vector(VECTOR x, VECTOR y, VECTOR z)
{
  int n,i;

  if (z == NULL) {
#ifdef CHECK_DIM
    if ((n = dim(x)) <= 0 || n != dim(y)) DIE_DIM;
#else
    n = dim(x);
#endif

    for (i=0; i<n; i++) x[i] -= y[i];
  }
  else {
#ifdef CHECK_DIM
    if ((n = dim(x)) <= 0 || n != dim(y) || n != dim(z)) DIE_DIM;
#else
    n = dim(x);
#endif

    for (i=0; i<n; i++) z[i] = x[i] - y[i];
  }
}

FLOAT_T prod_vector(VECTOR x, VECTOR y)
{
  int n,i;
  FLOAT_T p;

#ifdef CHECK_DIM
  if ((n = dim(x)) <= 0 || n != dim(y)) DIE_DIM;
#else
  n = dim(x);
#endif

  p = 0;
  for (i=0; i<n; i++) p += x[i] * y[i];

  return p;
}

/* [Note] Size of use0[] (use1[]) should be # of rows (columns) in m0 */
void proj_matrix(MATRIX m0, int *use0, int *use1, MATRIX m)
{
  int n0=-1,n1=-1,i,j,k,l;

#ifdef CHECK_DIM
  if ((n0 = dim0(m0)) <= 0 || (n1 = dim1(m0)) <= 0) DIE_DIM;
  k = 0;
  for (i=0; i<n0; i++) if (use0[i]) k++;
  if (k != dim0(m)) DIE_DIM;
  l = 0;
  for (j=0; j<n1; j++) if (use1[j]) l++;
  if (l != dim1(m)) DIE_DIM;
#else
  n0 = dim0(m0);
  n1 = dim1(m0);
#endif

  k = 0;
  for (i=0; i<n0; i++) {
    if (use0[i]) {
      l = 0;
      for (j=0; j<n1; j++) {
        if (use1[j]) {
          m[k][l] = m0[i][j];
          l++;
        }
      }
      k++;
    }
  }
}

/* [Note] Size of use[] should be the length of v0 */
void proj_vector(VECTOR v0, int *use, VECTOR v)
{
  int n,i,k;

#ifdef CHECK_DIM
  if ((n = dim(v0)) <= 0) DIE_DIM;
  k = 0;
  for (i=0; i<n; i++) if (use[i]) k++;
  if (k != dim(v)) DIE_DIM;
#else
  n = dim(v0);
#endif

  k = 0;
  for (i=0; i<n; i++) {
    if (use[i]) {
      v[k] = v0[i];
      k++;
    }
  }
}

/* Convert (n x 1)- or (1 x n)-matrix to n-dimension vector */
void matrix2vector(MATRIX m, VECTOR v)
{
  int n,i;

#ifdef CHECK_DIM
  if ((n = dim(v)) <= 0) DIE_DIM;
#else
  n = dim(v);
#endif

  if (dim0(m) == n && dim1(m) == 1) {
    for (i = 0; i < n; i++)
      v[i] = m[i][0];
  }
  else if (dim0(m) == 1 && dim1(m) == n) {
    memcpy(v,m[0],sizeof(FLOAT_T)*n);
  }
  else DIE_DIM;
}

/* Convert n-dimension vector to (n x 1)- or (1 x n)-matrix */
void vector2matrix(VECTOR v, MATRIX m)
{
  int n;

#ifdef CHECK_DIM
  if ((n = dim(v)) <= 0 ||
      !((dim0(m) == n && dim1(m) == 1) || (dim0(m) == 1 && dim1(m) == n)))
    DIE_DIM;
#else
  n = dim(v);
#endif

  /* Assumption: all elements in a matrix are allocated in a consective area */
  memcpy(m[0],v,sizeof(FLOAT_T)*n);
}

void mul_const_vector(VECTOR v, FLOAT_T c)
{
  int n,i;

#ifdef CHECK_DIM
  if ((n = dim(v)) <= 0) DIE_DIM;
#else
  n = dim(v);
#endif

  for (i=0; i<n; i++)
    v[i] *= c;
}

void div_const_vector(VECTOR v, FLOAT_T c)
{
  int n,i;

#ifdef CHECK_DIM
  if ((n = dim(v)) <= 0) DIE_DIM;
#else
  n = dim(v);
#endif

  for (i=0; i<n; i++)
    v[i] /= c;
}

void mul_const_matrix(MATRIX m, FLOAT_T c)
{
  int n0=-1,n1=-1,i,j;

#ifdef CHECK_DIM
  if ((n0 = dim0(m)) <= 0 || (n1 = dim1(m)) <= 0) DIE_DIM;
#else
  n0 = dim0(m);
  n1 = dim1(m);
#endif

  for (i=0; i<n0; i++)
    for (j=0; j<n1; j++)
      m[i][j] *= c;
}

void div_const_matrix(MATRIX m, FLOAT_T c)
{
  int n0=-1,n1=-1,i,j;

#ifdef CHECK_DIM
  if ((n0 = dim0(m)) <= 0 || (n1 = dim1(m)) <= 0) DIE_DIM;
#else
  n0 = dim0(m);
  n1 = dim1(m);
#endif

  for (i=0; i<n0; i++)
    for (j=0; j<n1; j++)
      m[i][j] /= c;
}

/* Returns a non-zero value if the input matrix is not valid */
int verify_matrix(MATRIX a)
{
  int n0=-1,n1=-1,i,j;

#ifdef CHECK_DIM
  if ((n0 = dim0(a)) <= 0 || (n1 = dim1(a)) <= 0) DIE_DIM;
#else
  n0 = dim0(a);
  n1 = dim1(a);
#endif

  for (i=0; i<n0; i++)
    for (j=0; j<n1; j++)
      if (ISNAN(a[i][j]) || ISINF(a[i][j])) return 1;

  return 0;
}

/* Returns a non-zero value if the input vector is not valid */
int verify_vector(VECTOR v)
{
  int n,i;

#ifdef CHECK_DIM
  if ((n = dim(v)) <= 0) DIE_DIM;
#else
  n = dim(v);
#endif

  for (i=0; i<n; i++)
    if (ISNAN(v[i]) || ISINF(v[i])) return 1;

  return 0;
}

/* Copy the elements in the upper triangle to the corresponding elements
 * in the lower triangle
 */
void copy_upper_to_lower(MATRIX m)
{
  int n,a0,a1;

#ifdef CHECK_DIM
  if ((n = dim0(m)) <= 0 || n != dim1(m)) DIE_DIM;
#else
  n = dim0(m);
#endif

  for (a0=0; a0<n; a0++)
    for (a1=0; a1<a0; a1++)
      m[a0][a1] = m[a1][a0];
}

void fprint_vector(FILE *fp, VECTOR v)
{
  int n,i;

#ifdef CHECK_DIM
  if ((n = dim(v)) <= 0) DIE_DIM;
#else
  n = dim(v);
#endif

  for (i=0; i<n; i++) {
    if (i == 0) fprintf(fp,"[");
    else fprintf(fp," ");
    fprintf(fp,conv_msg("%.Nf"),v[i]);
    if (i == (n - 1)) fprintf(fp,"   ]\n");
    else fprintf(fp,"; ");
  }
}

void fprint_matrix(FILE *fp, MATRIX a)
{
  int n0=-1,n1=-1,i,j;

#ifdef CHECK_DIM
  if ((n0 = dim0(a)) <= 0 || (n1 = dim1(a)) <= 0) DIE_DIM;
#else
  n0 = dim0(a);
  n1 = dim1(a);
#endif

  for (i=0; i<n0; i++) {
    if (i == 0) fprintf(fp,"[");
    else fprintf(fp," ");
    for (j=0; j<n1; j++)
      fprintf(fp,conv_msg("\t%.Nf"),a[i][j]);
    if (i == (n0 - 1)) fprintf(fp,"   ]\n");
    else fprintf(fp,";\n");
  }
}

void print_vector(VECTOR v)
{
  fprint_vector(stdout,v);
}

void print_matrix(MATRIX a)
{
  fprint_matrix(stdout,a);
}

void fprint_matrix_p(FILE *fp, MATRIX a, int *p)
{
  int n0,n1,i,j,i0;

  n0 = dim0(a);
  n1 = dim1(a);

  for (i=0; i<n0; i++) {
    if (i == 0) fprintf(fp,"[");
    else fprintf(fp," ");
    i0 = p[i];
    for (j=0; j<n1; j++)
      fprintf(fp,conv_msg("\t%.Ng"),a[i0][j]);
    if (i == (n0 - 1)) fprintf(fp,"   ]\n");
    else fprintf(fp,";\n");
  }
}

void print_matrix_p(MATRIX a, int *p)
{
  fprint_matrix_p(stdout,a,p);
}
