/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

#define DESC_ID_STR_SIZE 128
#define DESC_COL_STR_SIZE 128
#define DESC_COL_SIZE 10
#define DESC_COMM_STR_SIZE 1024
#define DESC_COMM_SIZE 3
#define DATA_COL_STR_SIZE (MAX_WORD_LEN + 1)

extern char version[];
extern char usage[];
extern char usage_full[];

static char desc_id[DESC_ID_STR_SIZE];
static char **desc_col;
static char **desc_comm;
static int desc_col_size;
static int desc_comm_size;
static int max_desc_col_size;
static int max_desc_comm_size;
static char data_line[MAX_OUT_LEN];
static int data_line_offset;

static void initialize_desc_col(void)
{
  int i;

  desc_col = (char **)xmalloc(sizeof(char *)*DESC_COL_SIZE);
  for (i=0; i<DESC_COL_SIZE; i++)
    desc_col[i] = alloc_str(DESC_COL_STR_SIZE);

  desc_col_size = 0;
  max_desc_col_size = DESC_COL_SIZE;
}

static void expand_desc_col(void)
{
  int i,old_size,new_size;

  old_size = max_desc_col_size;
  new_size = old_size * 2;

  desc_col = (char **)xrealloc(desc_col,sizeof(char *)*new_size);
  for (i=old_size; i<new_size; i++)
    desc_col[i] = alloc_str(DESC_COL_STR_SIZE);

  max_desc_col_size = new_size;
}

static void release_desc_col(void)
{
  int i;
  for (i=0; i<max_desc_col_size; i++)
    xfree(desc_col[i]);
  xfree(desc_col);
}

static void initialize_desc_comm(void)
{
  int i;

  desc_comm = (char **)xmalloc(sizeof(char *)*DESC_COMM_SIZE);
  for (i=0; i<DESC_COMM_SIZE; i++)
    desc_comm[i] = alloc_str(DESC_COMM_STR_SIZE);

  desc_comm_size = 0;
  max_desc_comm_size = DESC_COMM_SIZE;
}

static void expand_desc_comm(void)
{
  int i,old_size,new_size;

  old_size = max_desc_comm_size;
  new_size = old_size * 2;

  desc_comm = (char **)xrealloc(desc_comm,sizeof(char *)*new_size);
  for (i=old_size; i<new_size; i++)
    desc_comm[i] = alloc_str(DESC_COMM_STR_SIZE);

  max_desc_comm_size = new_size;
}

static void release_desc_comm(void)
{
  int i;
  for (i=0; i<max_desc_comm_size; i++)
    xfree(desc_comm[i]);
  xfree(desc_comm);
}

static void initialize_data_line(void)
{
  data_line_offset = 0;
  memset(data_line,0,MAX_OUT_LEN);
}

void initialize_printer(void)
{
  initialize_desc_col();
  initialize_desc_comm();
  initialize_data_line();
}

void release_printer(void)
{
  release_desc_col();
  release_desc_comm();
}

void print_version(FILE *fp)
{
  fputs(conv_msg(version),fp);
}

void print_usage(FILE *fp)
{
  NEWLINE(fp);
  print_version(fp);
  fputs(conv_msg(usage),fp);
  NEWLINE(fp);
}

void print_usage_full(FILE *fp)
{
  NEWLINE(fp);
  print_version(fp);
  fputs(conv_msg(usage_full),fp);
  NEWLINE(fp);
}

void add_desc_id(char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  xvsnprintf(desc_id,DESC_ID_STR_SIZE,conv_desc(fmt),ap);
  va_end(ap);
}

static void clean_desc(void)
{
  int i;

  desc_id[0] = '\0';
  for (i=0; i<desc_col_size;  i++) desc_col[i][0]  = '\0';
  for (i=0; i<desc_comm_size; i++) desc_comm[i][0] = '\0';
  desc_col_size = 0;
  desc_comm_size = 0;
}

static void clean_data(void)
{
  data_line_offset = 0;
  data_line[0] = '\0';
}

/* We do not call fflush() */
void flush_desc(FILE *fp)
{
  int i;

  fprintf(fp,"#### ID:%s",desc_id);
  NEWLINE(fp);

  if (desc_col_size > 0) {
    fputs("#### ",fp);
    for (i=0; i<desc_col_size; i++) {
      if (i > 0) fputs(", ",fp);
      fputs(desc_col[i],fp);
    }
    NEWLINE(fp);
  }

  if (desc_comm_size == 0) {
    fputs("####",fp);
    NEWLINE(fp);
  }
  else {
    for (i=0; i<desc_comm_size; i++) {
      fprintf(fp,"#### -- %s",desc_comm[i]);
      NEWLINE(fp);
    }
    fputs("####",fp);
    NEWLINE(fp);
  }

  clean_desc();
}

/* We do not call fflush() */
void flush_data(FILE *fp)
{
  char *q;

  if (data_line_offset <= 0) return;

  if (data_line_offset > MAX_OUT_LEN)
    die("Output buffer overflow (size >= %d)\n",MAX_OUT_LEN);

  q = data_line + data_line_offset;
  *q = '\0';

  fputs(data_line,fp);
  NEWLINE(fp);

  clean_data();
}

void print_data(FILE *fp, char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vfprintf(fp,conv_data(fmt),ap);
  va_end(ap);

  NEWLINE(fp);
}

void add_desc_col(char *fmt, ...)
{
  va_list ap;
  char *s;

  if (desc_col_size == max_desc_col_size) expand_desc_col();
  s = desc_col[desc_col_size];

  va_start(ap,fmt);
  xvsnprintf(s,DESC_COL_STR_SIZE,conv_desc(fmt),ap);
  va_end(ap);

  desc_col_size++;
}

void add_desc_comm(char *fmt, ...)
{
  va_list ap;
  char *s;

  if (desc_comm_size == max_desc_comm_size) expand_desc_comm();
  s = desc_comm[desc_comm_size];

  va_start(ap,fmt);
  xvsnprintf(s,DESC_COMM_STR_SIZE,conv_desc(fmt),ap);
  va_end(ap);

  desc_comm_size++;
}

/* Generic, but slow... */
void add_data(char *fmt, ...)
{
  static char *prec_out_str = NULL;
  va_list ap;
  int offset;

  if (prec_out_str == NULL) prec_out_str = int2strdup(prec_out);

  if (data_line_offset > 0) {
    data_line[data_line_offset] = (format_out == FMT_CSV) ? ',' : '\t';
    offset = data_line_offset + 1;
  }
  else offset = 0;

  va_start(ap,fmt);
  xvsnprintf((data_line + offset),
             DATA_COL_STR_SIZE,
             conv_base(prec_out_str,fmt),
             ap);
  va_end(ap);

  data_line_offset = strlen(data_line);
}

void add_data_str(char *s)
{
  int offset,len;

  if (data_line_offset > 0) {
    data_line[data_line_offset] = (format_out == FMT_CSV) ? ',' : '\t';
    offset = data_line_offset + 1;
  }
  else offset = 0;

  len = strlen(s);
  strcpy((data_line + offset),s);

  data_line_offset = offset + len;
}

void add_data_int(int i)
{
  int offset,len;
  char *s;

  if (data_line_offset > 0) {
    data_line[data_line_offset] = (format_out == FMT_CSV) ? ',' : '\t';
    offset = data_line_offset + 1;
  }
  else offset = 0;

  s = int2str(i);
  len = strlen(s);
  strcpy((data_line + offset),s);

  data_line_offset = offset + len;
}

void add_data_float(FLOAT_T f)
{
  int offset,len;
  char *s;

  if (data_line_offset > 0) {
    data_line[data_line_offset] = (format_out == FMT_CSV) ? ',' : '\t';
    offset = data_line_offset + 1;
  }
  else offset = 0;

  s = float2str(f);
  len = strlen(s);
  strcpy((data_line + offset),s);

  data_line_offset = offset + len;
}

/* [Note]
 * This function may return a string allocated in the static area.
 * So we should make a copy of the returned string if necessary
 * (e.g. using together with add_data()).
 */
char *get_case_id(int e)
{
  static char s[INT2STR_BUFSIZE];

  if (mode_id)
    return case_id_list[freq_list[e].case_id];
  else {
    strncpy(s,int2str(e),INT2STR_BUFSIZE);
    return s;
  }
}

char *get_ans_class(int e)
{
  return ans_class_list[freq_list[e].ans_id];
}

/* [Note]
 * This function may return a string allocated in the static area
 * of float2str().  So we should make a copy of the returned
 * string if necessary (e.g. using together with add_data()/add_data_str()).
 */
char *freq_entry_val(int e, int v)
{
  char *p = NULL;
  int u,a,v_d;

  if (is_attr_cont(v)) {
    u = attr_info[v].gauss_id;
    a = attr_info[v].local_id;

    if (freq_list[e].miss_g[u][a])
      p = "?";
    else
      p = float2str(freq_list[e].gval[u][a]);
  }
  else {
    v_d = attr_info[v].local_id;

    if (freq_list[e].id[v_d] >= 0)
      p = word_list[v_d][freq_list[e].id[v_d]];
    else if (freq_list[e].miss_w[v_d])
      p = "?";
    else die("Internal error: freq_entry_val(): word not assigned");
  }

  return p;
}

char *mode_str(int mode)
{
  return (mode > 0) ? "Enabled" : "Disabled";
}
