/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

static FLOAT_T **tmp_jmerge;
static FLOAT_T *tmp_jsplit;

int is_decreased_sm(int t, int m, int n)
{
  /* This would not happen without a bug or a numeric error */
  if (mode_daem &&
      log_itemp_sm[t][m] - log_itemp_sm[t][m-1] < ACCEPT_SMALL_NUMBER &&
      lambda_sm[t][m] < lambda_sm[t][m-1] &&
      (lambda_sm[t][m-1] - lambda_sm[t][m]) > ACCEPT_SMALL_NUMBER) {
    print_decrease_sm(fp_log,t,m,
                      lambda_sm[t][m-1],lambda_sm[t][m],log_itemp_sm[t][m]);
    print_param(fp_log);
    die(conv_msg("\n(iter:%d) LL or LPost decreased! "
                 "(before: %.Nf (%.Nf) | after: %.Nf (%.Nf) | diff: %.Nf) "
                 "-- see the log file\n"),
        m,lambda_sm[t][m-1],log_itemp_sm[t][m-1],
        lambda_sm[t][m],log_itemp_sm[t][m],
        (lambda_sm[t][m-1] - lambda_sm[t][m]));
    return 1;
  }
  else if (!mode_daem &&
           lambda_sm[t][m] < lambda_sm[t][m-1] &&
           (lambda_sm[t][m-1] - lambda_sm[t][m]) > ACCEPT_SMALL_NUMBER) {
    print_decrease_sm(fp_log,t,m,lambda_sm[t][m-1],lambda_sm[t][m],0.0);
    print_param(fp_log);
    die(conv_msg("\n(iter:%d) LL or LPost decreased! "
                 "(before: %.Nf | after: %.Nf | diff: %.Nf)"
                 " -- see the log file\n"),
        m,lambda_sm[t][m-1],lambda_sm[t][m],
        (lambda_sm[t][m-1] - lambda_sm[t][m]));
    return 1;
  }
  return 0;
}

int is_converged_sm(int t, int m, int n)
{
  if (m >= n) {
    print_msg(" (Stopped: %d iters)\tL=%.Nf\n",m,lambda_sm[t][m]);
    return 1;
  }
  else if (lambda_sm[t][m] - lambda_sm[t][m-1] > 0 &&
           lambda_sm[t][m] - lambda_sm[t][m-1] < epsilon) {
    print_msg(" (Converged: %d iters)\tL=%.Nf\n",m,lambda_sm[t][m]);
    return 1;
  }
  return 0;
}

/* m: total number of iters,
 * mm: number of iters at the current inverse temperature
 * itemp: should be =< 1.0
 */
int is_converged_sm_daem(int t, int m, int mm, int n, FLOAT_T itemp)
{
  if (itemp < 1.0) {
    if (mm >= n) {
      return 1;
    }
    else if (log_itemp_sm[t][m] - log_itemp_sm[t][m-1] < ACCEPT_SMALL_NUMBER &&
             lambda_sm[t][m] - lambda_sm[t][m-1] > 0 &&
             lambda_sm[t][m] - lambda_sm[t][m-1] < epsilon) {
      return 1;
    }
  }
  else {
    if (mm >= n) {
      print_msg(" (Stopped: %d iters)\tL=%.Nf\n",m,lambda_sm[t][m]);
      return 1;
    }
    else if (log_itemp_sm[t][m] - log_itemp_sm[t][m-1] < ACCEPT_SMALL_NUMBER &&
             lambda_sm[t][m] - lambda_sm[t][m-1] > 0 &&
             lambda_sm[t][m] - lambda_sm[t][m-1] < epsilon) {
      print_msg(" (Converged: %d iters)\tL=%.Nf\n",m,lambda_sm[t][m]);
      return 1;
    }
  }

  return 0;
}

/* returns the existence of singularity (including collinearity) */
static int is_singular_sm(int t, int m, int n)
{
  int k,u;

  for (k=0; k<num_class; k++) {
    for (u=0; u<gauss_size; u++) {
      if (gauss_attr_size[u] > 1 &&
          FABS(param_sdet[k][u]) < VERY_SMALL_FLOAT) {
        print_msg(" (Stopped: %d iters -- singular covars (det=%.Ne)"
                  " at class#=%d, gauss#=%d)\tL=%.Nf\n",
                  m,param_sdet[k][u],k,u,lambda_sm[t][m]);
        return 1;
      }
    }
  }

  return 0;
}

/* returns the existence of singularity (including collinearity) */
static int is_singular_part(int t, int m, int n, int *cand)
{
  int l,k,u;

  for (l=0; l<3; l++) {
    k = cand[l];

    for (u=0; u<gauss_size; u++) {
      if (gauss_attr_size[u] > 1 &&
          FABS(param_sdet[k][u]) < VERY_SMALL_FLOAT) {
        print_msg(" (Stopped: %d iters -- singular covars (det=%.Ne)"
                  " at class#=%d, gauss#=%d)\tL=%.Nf\n",
                  m,param_sdet[k][u],k,u,lambda_sm[t][m]);
        return 1;
      }
    }
  }

  return 0;
}

static int smem_iterate_full(int t, int m, int n, FLOAT_T itemp_init0)
{
  int mm;

  itemp = (mode_daem) ? itemp_init0 : 1.0;

  if (m == 0) {
    if (mode_daem && itemp < 1.0)
      compute_complete_prob_daem();
    else
      compute_complete_prob();
    compute_incomplete_prob();

    loglike_sm[t][m] = compute_loglike()/itemp;
    if (mode_smooth) {
      log_prior_sm[t][m] = compute_log_prior();
      lambda_sm[t][m] = loglike_sm[t][m] + log_prior_sm[t][m];
    }
    else {
      lambda_sm[t][m] = loglike_sm[t][m];
    }
    if (mode_daem) log_itemp_sm[t][m] = itemp;
  }

  while (1) { /* Loop for inverse temperatures */

    print_daem_message_sm(t,m,itemp);
    mm = 0;

    while (1) {
      print_dot_message(m);

      compute_conditional_prob();
      compute_expect();
      update_param();

      if (mode_gauss && is_singular_sm(t,m,n))
        break; /* just discard the last update */

      m++;
      mm++;

      if (mode_daem && itemp < 1.0)
        compute_complete_prob_daem();
      else
        compute_complete_prob();
      compute_incomplete_prob();

      if (m >= ll_sm_size[t]) expand_loglike_sm_all(t);
      loglike_sm[t][m] = compute_loglike()/itemp;
      if (mode_smooth) {
        log_prior_sm[t][m] = compute_log_prior();
        lambda_sm[t][m] = loglike_sm[t][m] + log_prior_sm[t][m];
      }
      else {
        lambda_sm[t][m] = loglike_sm[t][m];
      }
      if (mode_daem) log_itemp_sm[t][m] = itemp;

      if (is_decreased_sm(t,m,n)) break;

      if (mode_daem && is_converged_sm_daem(t,m,mm,n,itemp)) {
        break;
      }
      else if (!mode_daem && is_converged_sm(t,m,n)) {
        break;
      }
    }

    if (!mode_daem && m >= n) break;

    if (itemp < 1.0) itemp *= itemp_rate;
    else break;
    if (itemp > 1.0) itemp = 1.0;
  }

  final_loglike = loglike_sm[t][m];
  final_lambda = lambda_sm[t][m];
  if (mode_smooth) final_log_prior = log_prior_sm[t][m];

  return m;
}

int smem_iterate_part(int t, int h, int m, int n)
{
  int *cand;

  cand = sm_candidate[h];

  initialize_param_part(cand);
  save_param_part(cand);

  compute_complete_prob(); /* it is assumed itemp == 1 */
  compute_incomplete_prob();
  save_incomplete_prob_part(cand);

  loglike_sm[t][m] = compute_loglike();
  if (mode_smooth) {
    log_prior_sm[t][m] = compute_log_prior();
    lambda_sm[t][m] = loglike_sm[t][m] + log_prior_sm[t][m];
  }
  else {
    lambda_sm[t][m] = loglike_sm[t][m];
  }
  if (mode_daem) log_itemp_sm[t][m] = itemp;

  print_dot_message(m);

  while (1) {
    compute_conditional_prob_part(cand);
    compute_expect_part(cand);
    update_param_part(cand);

    if (mode_gauss && is_singular_part(t,m,n,cand))
      break;  /* just discard the last update */

    m++;

    compute_complete_prob_part(cand);
    compute_incomplete_prob_part(cand);

    loglike_sm[t][m] = compute_loglike();
    if (mode_smooth) {
      log_prior_sm[t][m] = compute_log_prior();
      lambda_sm[t][m] = loglike_sm[t][m] + log_prior_sm[t][m];
    }
    else {
      lambda_sm[t][m] = loglike_sm[t][m];
    }
    if (mode_daem) log_itemp_sm[t][m] = itemp;

    print_dot_message(m);

    if (lambda_sm[t][m] < lambda_sm[t][m-1] &&
        lambda_sm[t][m-1] - lambda_sm[t][m] > ACCEPT_SMALL_NUMBER) {
      print_msg("\n(part) (iter:%d) LL or LPost decreased!"
                " (curr: %.Nf / last: %.Nf)\n",
                m,lambda_sm[t][m],lambda_sm[t][m-1]);
      break;
    }

    if (m >= n) {
      print_msg("(S:%d)",m);
      break;
    }
    else {
      if (lambda_sm[t][m] - lambda_sm[t][m-1] < epsilon) {
        print_msg("(C:%d)",m);
        break;
      }
    }
  }

  /* do not update final_loglike, etc. */

  return m;
}

static void check_jmerge(int k1, int k2)
{
  if (ISNAN(jmerge[k1][k2])) {
    print_param(fp_log);
    DIE_NUMERIC("Computed Jmerge is not a number -- see the log file\n");
  }
  if (ISINF(jmerge[k1][k2])) {
    print_param(fp_log);
    DIE_NUMERIC("Computed Jmerge is not finite -- see the log file\n");
  }
}

void compute_jmerge()
{
  int k,k1,k2,e;
  FLOAT_T *norm;

  if (mode_norm) {
    norm = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_class);

#pragma omp parallel for private(e)
    for (k=0; k<num_class; k++) {
      norm[k] = 0.0;
      for (e=0; e<num_entry; e++)
        norm[k] += freq_list[e].freq * prob_cond[k][e] * prob_cond[k][e];
    }

#pragma omp parallel for private(k2,e)
    for (k1=0; k1<num_class; k1++) {
      for (k2=(k1+1); k2<num_class; k2++) {
        jmerge[k1][k2] = 0.0;
        for (e=0; e<num_entry; e++)
          jmerge[k1][k2] +=
            freq_list[e].freq * prob_cond[k1][e] * prob_cond[k2][e];
        jmerge[k1][k2] /= SQRT(norm[k1]*norm[k2]);
        check_jmerge(k1,k2);
      }
    }

    xfree(norm);
  }
  else {
#pragma omp parallel for private(k2,e)
    for (k1=0; k1<num_class; k1++) {
      for (k2=(k1+1); k2<num_class; k2++) {
        jmerge[k1][k2] = 0.0;
        for (e=0; e<num_entry; e++)
          jmerge[k1][k2] +=
            freq_list[e].freq * prob_cond[k1][e] * prob_cond[k2][e];
        check_jmerge(k1,k2);
      }
    }
  }

}

static int compare_candidate(const int **x, const int **y)
{
  int k1_x = (*x)[0];
  int k2_x = (*x)[1];
  int k3_x = (*x)[2];
  int k1_y = (*y)[0];
  int k2_y = (*y)[1];
  int k3_y = (*y)[2];

  FLOAT_T jm_x = tmp_jmerge[k1_x][k2_x];
  FLOAT_T jm_y = tmp_jmerge[k1_y][k2_y];
  FLOAT_T js_x = tmp_jsplit[k3_x];
  FLOAT_T js_y = tmp_jsplit[k3_y];

  if (k1_x == k1_y && k2_x == k2_y) {
    if (js_x > js_y) return -1;
    if (js_x < js_y) return 1;
    return 0;
  }
  else {
    if (jm_x > jm_y) return -1;
    if (jm_x < jm_y) return 1;
    return 0;
  }
}

void sort_candidate()
{
  int k1,k2,k3;
  int h;

  h = 0;
  for (k1=0; k1<num_class; k1++)
    for (k2=(k1+1); k2<num_class; k2++)
      for (k3=0; k3<num_class; k3++) {
        if (k3 == k1 || k3 == k2) continue;
        sm_candidate[h][0] = k1;
        sm_candidate[h][1] = k2;
        sm_candidate[h][2] = k3;
        h++;
      }

  tmp_jmerge = jmerge;
  tmp_jsplit = jsplit;

  qsort(sm_candidate,num_candidate,sizeof(int *),
        (int(*)(const void *, const void *))compare_candidate);
}

static void compute_incomplete_prob_part_log(int *cand)
{
  int l,k,e;
  FLOAT_T sum,pcomp_max;

  if (num_class > 3) {
#pragma omp parallel for private(l,k,sum)
    for (e=0; e<num_entry; e++) {
      sum = 1.0;  /* not 0.0! */
      for (l=0; l<3; l++) {
        k = cand[l];
        sum += EXP(prob_comp[k][e] - prob_incomp_part[e]);
      }

      prob_incomp[e] = prob_incomp_part[e] + LOG(sum);
    }
  }
  else {
#pragma omp parallel for private(l,k,pcomp_max,sum)
    for (e=0; e<num_entry; e++) {
      pcomp_max = 0.0;
      for (l=0; l<3; l++) {
        k = cand[l];
        if (l == 0 || prob_comp[k][e] > pcomp_max)
          pcomp_max = prob_comp[k][e];
      }

      sum = 0.0;
      for (l=0; l<3; l++) {
        k = cand[l];
        sum += EXP(prob_comp[k][e] - pcomp_max);
      }

      prob_incomp[e] = pcomp_max + LOG(sum);
    }
  }
}

static void compute_incomplete_prob_part_nolog(int *cand)
{
  int l,k,e;

#pragma omp parallel for private(l,k)
  for (e=0; e<num_entry; e++) {
    prob_incomp[e] = (num_class > 3) ? prob_incomp_part[e] : 0;
    for (l=0; l<3; l++) {
      k = cand[l];
      prob_incomp[e] += prob_comp[k][e];
    }
  }
}

void compute_incomplete_prob_part(int *cand)
{
  if (mode_log_scale && mode_smooth_strict)
    compute_incomplete_prob_part_log(cand);
  else
    compute_incomplete_prob_part_nolog(cand);
}

static void compute_conditional_prob_part_log(int *cand)
{
  int l,k,e;

  for (l=0; l<3; l++) {
    k = cand[l];
#pragma omp parallel for
    for (e=0; e<num_entry; e++)
      prob_cond[k][e] = EXP(prob_comp[k][e] - prob_incomp[e]);
  }
}

static void compute_conditional_prob_part_nolog(int *cand)
{
  int l,k;
  int e;
  FLOAT_T p;

  for (l=0; l<3; l++) {
    k = cand[l];
#pragma omp parallel for private(p)
    for (e=0; e<num_entry; e++) {
      p = prob_incomp[e];
      if (p > TINY_FLOAT * TINY_FLOAT) {
        prob_cond[k][e] = prob_comp[k][e]/p;
      }
      else
        prob_cond[k][e] = 0.0;
    }
  }
}

void compute_conditional_prob_part(int *cand)
{
  if (mode_log_scale && mode_smooth_strict)
    compute_conditional_prob_part_log(cand);
  else
    compute_conditional_prob_part_nolog(cand);
}

static void save_incomplete_prob_part_log(int *cand)
{
  int k,e;
  int k1,k2,k3;
  int first;
  FLOAT_T pcomp_max,sum;

  if (num_class > 3) {
    k1 = cand[0];
    k2 = cand[1];
    k3 = cand[2];

    /* Cannot immediately add a `pragma omp' statement here */
    for (e=0; e<num_entry; e++) {
      pcomp_max = 0.0;
      first = 1;
      for (k=0; k<num_class; k++) {
        if (k == k1 || k == k2 || k == k3) continue;
        if (first || prob_comp[k][e] > pcomp_max) {
          pcomp_max = prob_comp[k][e];
          first = 0;
        }
      }

      sum = 0.0;
      for (k=0; k<num_class; k++) {
        if (k == k1 || k == k2 || k == k3) continue;
        sum += EXP(prob_comp[k][e] - pcomp_max);
      }

      prob_incomp_part[e] = pcomp_max + LOG(sum);
    }
  }
}

static void save_incomplete_prob_part_nolog(int *cand)
{
  int k,e;
  int k1,k2,k3;

  if (num_class > 3) {
    k1 = cand[0];
    k2 = cand[1];
    k3 = cand[2];

#pragma omp parallel for
    for (e=0; e<num_entry; e++)
      prob_incomp_part[e] = 0.0;

    /* Cannot immediately add a `pragma omp' statement here */
    for (k=0; k<num_class; k++) {
      if (k == k1 || k == k2 || k == k3) continue;
      for (e=0; e<num_entry; e++)
        prob_incomp_part[e] += prob_comp[k][e];
    }
  }
}

void save_incomplete_prob_part(int *cand)
{
  if (mode_log_scale && mode_smooth_strict)
    save_incomplete_prob_part_log(cand);
  else
    save_incomplete_prob_part_nolog(cand);
}

void save_conditional_prob_part(int *cand)
{
  int l,k,e;

  for (l=0; l<3; l++) {
    k = cand[l];
#pragma omp parallel for
    for (e=0; e<num_entry; e++)
      prob_cond_part[l][e] = prob_cond[k][e];
  }
}

void save_loglike_best_sm(int t,int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++) {
    loglike_best[m] = loglike_sm[t][m];
    lambda_best[m] = lambda_sm[t][m];
  }

  if (mode_smooth)
    for (m=0; m<num_iterate; m++) {
      log_prior_best[m] = log_prior_sm[t][m];
    }
}

void restore_loglike_best_sm(int t,int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++) {
    loglike_sm[t][m] = loglike_best[m];
    lambda_sm[t][m] = lambda_best[m];
  }

  if (mode_smooth)
    for (m=0; m<num_iterate; m++)
      log_prior_sm[t][m] = log_prior_best[m];
}

void save_log_itemp_best_sm(int t,int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++)
    log_itemp_best[m] = log_itemp_sm[t][m];
}

void restore_log_itemp_best_sm(int t,int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++)
    log_itemp_sm[t][m] = log_itemp_best[m];
}

static void compute_param_precisions_part(int *cand)
{
  int l,k,u;

#pragma omp parallel for private(k,u)
  for (l=0; l<3; l++) {
    k = cand[l];
    for (u=0; u<gauss_size; u++) {
      param_sdet[k][u] = INVERT(param_s[k][u],param_p[k][u]);
      if (err_inv) DIE_NUMERIC("Failed to invert covariances\n");
    }
  }
}

void initialize_param_part(int *cand)
{
  int i,k1,k2,k3,v;
  FLOAT_T r1,r2;
  FLOAT_T noise1,noise2;
  FLOAT_T sum1,sum2,sum3;
  FLOAT_T p1,p2,p3,u1,u2,u3,s1,s2;
  FLOAT_T diag,det;
  int u,n,a,a0,a1;

  k1 = cand[0];
  k2 = cand[1];
  k3 = cand[2];

  param_c[k1] = param_c[k1] + param_c[k2];
  param_c[k2] = 0.5 * param_c[k3];
  param_c[k3] = 0.5 * param_c[k3];

  r1 = expect_c[k1]/(expect_c[k1] + expect_c[k2]);
  r2 = expect_c[k2]/(expect_c[k1] + expect_c[k2]);

  /* To keep the repeatability, we do not use OpenMP here */
  if (mode_nominal) {
    for (v=0; v<tuple_size_d; v++) {
      sum1 = 0.0;
      sum2 = 0.0;
      sum3 = 0.0;
      for (i=0; i<num_word[v]; i++) {
        p1 = param_w[k1][v][i];
        p2 = param_w[k2][v][i];
        p3 = param_w[k3][v][i];
        noise1 = noise_ratio * random_float() / num_word[v];
        noise2 = noise_ratio * random_float() / num_word[v];
        param_w[k1][v][i] = r1 * p1 + r2 * p2;
        param_w[k2][v][i] = p3 + noise1;
        param_w[k3][v][i] = p3 + noise2;
        sum1 += param_w[k1][v][i];
        sum2 += param_w[k2][v][i];
        sum3 += param_w[k3][v][i];
      }
      /* normalization */
      for (i=0; i<num_word[v]; i++) {
        param_w[k1][v][i] /= sum1;
        param_w[k2][v][i] /= sum2;
        param_w[k3][v][i] /= sum3;
      }
    }
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];
      for (a=0; a<n; a++) {
        u1 = param_u[k1][u][a];
        u2 = param_u[k2][u][a];
        u3 = param_u[k3][u][a];
        noise1 =
          noise_ratio_gauss * SQRT(sample_s[u][a][a]) * (0.5 - random_float());
        noise2 =
          noise_ratio_gauss * SQRT(sample_s[u][a][a]) * (0.5 - random_float());
        param_u[k1][u][a] = r1 * u1 + r2 * u2;
        param_u[k2][u][a] = u3 + noise1;
        param_u[k3][u][a] = u3 + noise2;
      }

      det = DETERMINANT(param_s[k3][u]);
      if (err_det)
        DIE_NUMERIC("Failed to compute the determinant of covariances\n");
      diag = POW(det,1.0/n);

      for (a0=0; a0<n; a0++) {
        for (a1=a0; a1<n; a1++) {
          s1 = param_s[k1][u][a0][a1];
          s2 = param_s[k2][u][a0][a1];
          param_s[k1][u][a0][a1] = r1 * s1 + r2 * s2;
          param_s[k2][u][a0][a1] = (a0==a1) ? diag : 0;
          param_s[k3][u][a0][a1] = (a0==a1) ? diag : 0;
        }
      }
      copy_upper_to_lower(param_s[k1][u]);
      copy_upper_to_lower(param_s[k2][u]);
      copy_upper_to_lower(param_s[k3][u]);
    }
  }

  if (mode_gauss) compute_param_precisions_part(cand);
}

static void check_jsplit(int k)
{
  if (ISNAN(jsplit[k])) {
    print_param(fp_log);
    DIE_NUMERIC("Computed Jsplit[%d] is not a number -- see the log file\n");
  }
  if (ISINF(jsplit[k])) {
    print_param(fp_log);
    DIE_NUMERIC("Computed Jsplit[%d] is not finite -- see the log file\n");
  }
}

static void compute_jsplit_nolog(void)
{
  int e,k,v,u,n;
  FLOAT_T p_emp,p_x;
  VECTOR x,p_u;
  MATRIX p_s,p_p;
  int *obs;
  int n_miss;
  FLOAT_T det,p_lim;

  p_lim = TINY_FLOAT * TINY_FLOAT;

#pragma omp parallel for private(e,p_emp,v,p_x,u,n,p_u,p_s,p_p,det,x,n_miss,obs)
  for (k=0; k<num_class; k++) {
    jsplit[k] = 0.0;

    for (e=0; e<num_entry; e++) {
      p_emp = freq_list[e].freq * prob_cond[k][e] / expect_c[k];
      if (p_emp < p_lim) p_emp = p_lim;

      p_x = 1.0;
      if (mode_nominal) {
        if (total_num_miss_w > 0 && freq_list[e].num_miss_w > 0) {
          for (v=0; v<tuple_size_d; v++) {
            if (!freq_list[e].miss_w[v])
              p_x *= param_w[k][v][freq_list[e].id[v]];
          }
        }
        else {
          for (v=0; v<tuple_size_d; v++)
            p_x *= param_w[k][v][freq_list[e].id[v]];
        }
      }
      if (mode_gauss) {
        for (u=0; u<gauss_size; u++) {
          n = gauss_attr_size[u];

          p_u = param_u[k][u];
          p_s = param_s[k][u];
          p_p = param_p[k][u];
          det = param_sdet[k][u];

          x = freq_list[e].gval[u];
          n_miss = freq_list[e].num_miss_g[u];

          if (0 < n_miss && n_miss < n) {
            obs = freq_list[e].obs_g[u];
            p_x *= gauss_marginal(x,p_u,p_s,obs);
          }
          else if (n_miss == 0) {
            p_x *= gauss_prec(x,p_u,p_s,p_p,det);
          }
        }
      }

      if (p_x < p_lim) p_x = p_lim;

      jsplit[k] += p_emp * (LOG(p_emp) - LOG(p_x));
    }

    check_jsplit(k);
  }
}

static void compute_jsplit_log(void)
{
  int e,k,v,u,n;
  FLOAT_T p_emp,p_x;
  VECTOR x,p_u;
  MATRIX p_s,p_p;
  int *obs;
  int n_miss;
  FLOAT_T det,p_lim;

  p_lim = TINY_FLOAT * TINY_FLOAT;

#pragma omp parallel for private(e,p_emp,v,p_x,u,n,p_u,p_s,p_p,det,x,n_miss,obs)
  for (k=0; k<num_class; k++) {
    jsplit[k] = 0.0;

    for (e=0; e<num_entry; e++) {
      p_emp = freq_list[e].freq * prob_cond[k][e] / expect_c[k];
      if (p_emp < p_lim) p_emp = p_lim;

      p_x = 0.0;
      if (mode_nominal) {
        if (total_num_miss_w > 0 && freq_list[e].num_miss_w > 0) {
          for (v=0; v<tuple_size_d; v++) {
            if (!freq_list[e].miss_w[v])
              p_x += LOG(param_w[k][v][freq_list[e].id[v]]);
          }
        }
        else {
          for (v=0; v<tuple_size_d; v++)
            p_x += LOG(param_w[k][v][freq_list[e].id[v]]);
        }
      }
      if (mode_gauss) {
        for (u=0; u<gauss_size; u++) {
          n = gauss_attr_size[u];

          p_u = param_u[k][u];
          p_s = param_s[k][u];
          p_p = param_p[k][u];
          det = param_sdet[k][u];

          x = freq_list[e].gval[u];
          n_miss = freq_list[e].num_miss_g[u];

          if (0 < n_miss && n_miss < n) {
            obs = freq_list[e].obs_g[u];
            p_x += lgauss_marginal(x,p_u,p_s,obs);
          }
          else if (n_miss == 0) {
            p_x += lgauss_prec(x,p_u,p_s,p_p,det);
          }
        }
      }

      if (p_x < LOG(p_lim)) p_x = LOG(p_lim);

      jsplit[k] += p_emp * (LOG(p_emp) - p_x);
    }

    check_jsplit(k);
  }
}

void compute_jsplit(void)
{
  if (mode_log_scale && mode_smooth_strict)
    compute_jsplit_log();
  else
    compute_jsplit_nolog();
}

static void compute_complete_prob_part_log(int *cand)
{
  int l,k,v,e,u,n;
  VECTOR x,p_u;
  MATRIX p_s,p_p;
  int *obs;
  int n_miss;
  FLOAT_T det;

#pragma omp parallel for private(k,e)
  for (l=0; l<3; l++) {
    k = cand[l];
    for (e=0; e<num_entry; e++)
      prob_comp[k][e] = LOG(param_c[k]);
  }

  if (mode_nominal) {
    for (l=0; l<3; l++) {
      k = cand[l];
#pragma omp parallel for private(v)
      for (e=0; e<num_entry; e++) {
        if (total_num_miss_w > 0 && freq_list[e].num_miss_w > 0) {
          for (v=0; v<tuple_size_d; v++)
            if (!freq_list[e].miss_w[v])
              prob_comp[k][e] += LOG(param_w[k][v][freq_list[e].id[v]]);
        }
        else {
          for (v=0; v<tuple_size_d; v++)
            prob_comp[k][e] += LOG(param_w[k][v][freq_list[e].id[v]]);
        }
      }
    }
  }

  if (mode_gauss) {
#pragma omp parallel for private(k,u,n,p_u,p_s,p_p,det,x,e,n_miss,obs)
    for (l=0; l<3; l++) {
      k = cand[l];
      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];

        p_u = param_u[k][u];
        p_s = param_s[k][u];
        p_p = param_p[k][u];
        det = param_sdet[k][u];

        for (e=0; e<num_entry; e++) {
          x = freq_list[e].gval[u];
          n_miss = freq_list[e].num_miss_g[u];

          if (0 < n_miss && n_miss < n) {
            obs = freq_list[e].obs_g[u];
            prob_comp[k][e] += lgauss_marginal(x,p_u,p_s,obs);
          }
          else if (n_miss == 0) {
            prob_comp[k][e] += lgauss_prec(x,p_u,p_s,p_p,det);
          }
          else {
            /* if n_miss == n, do nothing */
          }
        }
      }
    }
  }
}

static void compute_complete_prob_part_nolog(int *cand)
{
  int l,k,v,e,u,n;
  VECTOR x,p_u;
  MATRIX p_s,p_p;
  int *obs;
  int n_miss;
  FLOAT_T det;

  for (l=0; l<3; l++) {
    k = cand[l];
    for (e=0; e<num_entry; e++)
      prob_comp[k][e] = param_c[k];
  }

  if (mode_nominal) {
    for (l=0; l<3; l++) {
      k = cand[l];
#pragma omp parallel for private(v)
      for (e=0; e<num_entry; e++) {
        if (total_num_miss_w > 0 && freq_list[e].num_miss_w > 0) {
          for (v=0; v<tuple_size_d; v++) {
            if (!freq_list[e].miss_w[v])
              prob_comp[k][e] *= param_w[k][v][freq_list[e].id[v]];
          }
        }
        else {
          for (v=0; v<tuple_size_d; v++)
            prob_comp[k][e] *= param_w[k][v][freq_list[e].id[v]];
        }
      }
    }
  }

  if (mode_gauss) {
#pragma omp parallel for private(k,u,n,p_u,p_s,p_p,det,x,e,n_miss,obs)
    for (l=0; l<3; l++) {
      k = cand[l];
      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];

        p_u = param_u[k][u];
        p_s = param_s[k][u];
        p_p = param_p[k][u];
        det = param_sdet[k][u];

        for (e=0; e<num_entry; e++) {
          x = freq_list[e].gval[u];
          n_miss = freq_list[e].num_miss_g[u];

          if (0 < n_miss && n_miss < n) {
            obs = freq_list[e].obs_g[u];
            prob_comp[k][e] *= gauss_marginal(x,p_u,p_s,obs);
          }
          else if (n_miss == 0) {
            prob_comp[k][e] *= gauss_prec(x,p_u,p_s,p_p,det);
          }
          else {
            /* if n_miss == n, do nothing */
          }
        }
      }
    }
  }
}

void compute_complete_prob_part(int *cand)
{
  if (mode_log_scale && mode_smooth_strict)
    compute_complete_prob_part_log(cand);
  else
    compute_complete_prob_part_nolog(cand);
}

static void compute_expect_gauss_part(int *cand)
{
  int l,k,e,u;

#pragma omp parallel for private(k,u)
  for (l=0; l<3; l++) {
    k = cand[l];
    for (u=0; u<gauss_size; u++) {
      clear_vector(expect_u[k][u]);
      clear_matrix(expect_s[k][u]);
    }
  }

#pragma omp parallel for private(k,e)
  for (l=0; l<3; l++) {
    k = cand[l];
    for (e=0; e<num_entry; e++)
      compute_expect_gauss0(k,e,freq_list[e].freq * prob_cond[k][e]);
  }
}

void compute_expect_part(int *cand)
{
  int l,k,i,e,v;
  FLOAT_T tmp_e;
  FREQ_PTR fp;
  FLOAT_T e_c_k;

  for (l=0; l<3; l++) {
    k = cand[l];
    e_c_k = 0.0;
#pragma omp parallel for reduction(+: e_c_k)
    for (e=0; e<num_entry; e++)
      e_c_k += freq_list[e].freq * prob_cond[k][e];
    expect_c[k] = e_c_k;
  }

  if (mode_nominal) {
    for (l=0; l<3; l++) {
      k = cand[l];

#pragma omp parallel for private(i)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          expect_w[k][v][i] = 0.0;

      /* Cannot immediately add a `pragma omp' statement here */
      for (e=0; e<num_entry; e++) {
        fp = &freq_list[e];
        tmp_e = fp->freq * prob_cond[k][e];
        for (v=0; v<tuple_size_d; v++) {
          if (!fp->miss_w[v]) {
            expect_w[k][v][fp->id[v]] += tmp_e;
          }
          else {
            for (i=0; i<num_word[v]; i++)
              expect_w[k][v][i] += tmp_e * param_w[k][v][i];
          }
        }
      }
    }
  }

  if (mode_gauss) compute_expect_gauss_part(cand);
}

void check_expect_part(int *cand)
{
  int k,l,i,v;
  FLOAT_T sum_c, sum_w;

  sum_c = 0.0;
  for (l=0; l<3; l++) {
    k = cand[l];
    sum_c += expect_c[k];
  }
  print_msg("check> (part) sum_E_c = %.Nf\n",sum_c);

  if (mode_nominal) {
    for (l=0; l<3; l++) {
      k = cand[l];
      for (v=0; v<tuple_size_d; v++) {
        sum_w = 0.0;
        for (i=0; i<num_word[v]; i++)
          sum_w += expect_w[k][v][i];
        print_msg("check> (part) sum_E_w[%d][%d] = %.Nf\n",k,v,sum_w);
      }
    }
  }
}

/* implemented for the future use */
void save_param_part(int *cand)
{
  int l,k;

  for (l=0; l<3; l++) {
    k = cand[l];
    param_c_part[l] = param_c[k];
  }
}

void update_param_part(int *cand)
{
  FLOAT_T denom_w;
  int k,l,i,v;
  int u,n,a,a0,a1;
  FLOAT_T param_c_sum0,param_c_sum,alpha_w0;
  FLOAT_T tmp_s;
  FLOAT_T smooth_u0,smooth_u1;

  /* compute the sum of parameters of a SM candidate */
  param_c_sum0 = 0.0;
  for (l=0; l<3; l++)
    param_c_sum0 += param_c_part[l];

  if (mode_read_hparam) {
    param_c_sum = 0.0;
    for (l=0; l<3; l++) {
      k = cand[l];
      param_c[k] = (expect_c[k] + hparam_c[k] - 1.0);
      param_c_sum += param_c[k];
    }
    for (l=0; l<3; l++) {
      k = cand[l];
      param_c[k] = param_c_sum0 * param_c[k] / param_c_sum;
    }
  }
  else if (smooth_c > 0) {
    param_c_sum = 0.0;
    for (l=0; l<3; l++) {
      k = cand[l];
      param_c[k] = (expect_c[k] + smooth_c);
      param_c_sum += param_c[k];
    }
    for (l=0; l<3; l++) {
      k = cand[l];
      param_c[k] = param_c_sum0 * param_c[k] / param_c_sum;
    }
  }
  else {
    param_c_sum = 0.0;
    for (l=0; l<3; l++) {
      k = cand[l];
      param_c_sum += expect_c[k];
    }
    for (l=0; l<3; l++) {
      k = cand[l];
      param_c[k] = param_c_sum0 * expect_c[k] / param_c_sum;
    }
  }

  if (mode_read_hparam) {
    for (l=0; l<3; l++) {
      k = cand[l];
#pragma omp parallel for private(denom_w,i)
      for (v=0; v<tuple_size_d; v++) {
        denom_w = 0.0;
        for (i=0; i<num_word[v]; i++)
          denom_w +=
            (param_w[k][v][i] = expect_w[k][v][i] + hparam_w[k][v][i] - 1.0);
        for (i=0; i<num_word[v]; i++)
          param_w[k][v][i] /= denom_w;
      }
    }
  }
  else {
#pragma omp parallel for private(k,l,denom_w,i)
    for (v=0; v<tuple_size_d; v++) {
      if (smooth_w[v] > 0) {
        for (l=0; l<3; l++) {
          k = cand[l];
          denom_w = expect_c[k] + smooth_w[v] * num_word[v];
          for (i=0; i<num_word[v]; i++)
            param_w[k][v][i] = (expect_w[k][v][i] + smooth_w[v])/denom_w;
        }
      }
      else {
        for (l=0; l<3; l++) {
          k = cand[l];
          denom_w = expect_c[k];
          for (i=0; i<num_word[v]; i++)
            param_w[k][v][i] = expect_w[k][v][i]/denom_w;
        }
      }
    }
  }

  if (mode_gauss) {
    if (mode_smooth_gauss) {
      for (l=0; l<3; l++) {
        k = cand[l];
        for (u=0; u<gauss_size; u++) {
          n = gauss_attr_size[u];

          alpha_w0 = get_alpha_w(k,u,n);

          for (a=0; a<n; a++) {
            smooth_u0 = hparam_u[k][u][a];
            param_u[k][u][a] =
              (tau * smooth_u0 + expect_u[k][u][a]) / (tau + expect_c[k]);
          }

          for (a0=0; a0<n; a0++) {
            smooth_u0 = hparam_u[k][u][a0];
            for (a1=a0; a1<n; a1++) {
              smooth_u1 = hparam_u[k][u][a1];

              tmp_s =
                (hparam_s[k][u][a0][a1]
                 + expect_s[k][u][a0][a1]
                 - param_u[k][u][a0] * expect_u[k][u][a1]
                 - param_u[k][u][a1] * expect_u[k][u][a0]
                 + expect_c[k] * param_u[k][u][a0] * param_u[k][u][a1]
                 + tau * (smooth_u0 - param_u[k][u][a0]) * (smooth_u1 - param_u[k][u][a1]))
                / ((alpha_w0 - n) + expect_c[k]);

              param_s[k][u][a0][a1] =
                (FABS(tmp_s) > SIGMA_LIMIT) ?
                tmp_s : ((a0 != a1 && tmp_s < 0) ? -SIGMA_LIMIT : SIGMA_LIMIT);
            }
          }
          copy_upper_to_lower(param_s[k][u]);
        }
      }
    }
    else {
#pragma omp parallel for private(k,u,n,a,a0,a1,tmp_s)
      for (l=0; l<3; l++) {
        k = cand[l];
        for (u=0; u<gauss_size; u++) {
          n = gauss_attr_size[u];

          for (a=0; a<n; a++)
            param_u[k][u][a] = expect_u[k][u][a] / expect_c[k];

          for (a0=0; a0<n; a0++) {
            for (a1=a0; a1<n; a1++) {
              tmp_s =
                (expect_s[k][u][a0][a1] / expect_c[k]) -
                param_u[k][u][a0] * param_u[k][u][a1];
              param_s[k][u][a0][a1] =
                (FABS(tmp_s) > SIGMA_LIMIT) ?
                tmp_s : ((a0 != a1 && tmp_s < 0) ? -SIGMA_LIMIT : SIGMA_LIMIT);
            }
          }
          copy_upper_to_lower(param_s[k][u]);
        }
      }
    }
  }

  if (mode_gauss) compute_param_precisions_part(cand);
}

void check_param_part(int *cand)
{
  int k,l,i,v;
  FLOAT_T sum_c, sum_w;

  sum_c = 0.0;
  for (l=0; l<3; l++) {
    k = cand[l];
    sum_c += param_c[k];
  }
  print_msg("check> (part) sum_c = %.Nf\n",sum_c);

  for (l=0; l<3; l++) {
    k = cand[l];
    for (v=0; v<tuple_size_d; v++) {
      sum_w = 0.0;
      for (i=0; i<num_word[v]; i++)
        sum_w += param_w[k][v][i];
      print_msg("check> (part) sum_w[%d][%d] = %.Nf\n",k,v,sum_w);
    }
  }
}

void save_param_sm(void)
{
  int k,v,i,u,n,a,a0,a1;

  for (k=0; k<num_class; k++)
    param_c_sm[k] = param_c[k];

  if (mode_nominal) {
#pragma omp parallel for private(v,i)
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          param_w_sm[k][v][i] = param_w[k][v][i];
    }
  }

  if (mode_gauss) {
#pragma omp parallel for private(u,a,a0,a1,n)
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];

        for (a=0; a<n; a++)
          param_u_sm[k][u][a] = param_u[k][u][a];

        for (a0=0; a0<n; a0++)
          for (a1=0; a1<n; a1++)
            param_s_sm[k][u][a0][a1] = param_s[k][u][a0][a1];
      }
    }
  }
}

void restore_param_sm(void)
{
  int k,v,i,u;

  for (k=0; k<num_class; k++)
    param_c[k] = param_c_sm[k];

  if (mode_nominal) {
#pragma omp parallel for private(v,i)
    for (k=0; k<num_class; k++) {
      param_c[k] = param_c_sm[k];
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          param_w[k][v][i] = param_w_sm[k][v][i];
    }
  }

  if (mode_gauss) {
#pragma omp parallel for private(u)
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        copy_vector(param_u[k][u],param_u_sm[k][u]);
        copy_matrix(param_s[k][u],param_s_sm[k][u]);
        param_sdet[k][u] = INVERT(param_s[k][u],param_p[k][u]);
        if (err_inv)
          DIE_NUMERIC("Failed to invert covariances\n");
      }
    }
  }
}

/*
 * SMEM run(s) with some paticular number of classes
 * -- u is the index for searching the optimal number of classes
 */
void smem_iterate(int u)
{
  int t,h;
  FLOAT_T last_lambda;
  FLOAT_T curr_jsplit, curr_jmerge;
  int sm_ok;
  int n,n0;
  int m1,m2;
  int k1,k2,k3,last_k1 = -1,last_k2 = -1;
  int num_tried_cand, num_tried_jsplit_cand = -1;
  int num_iterate,num_iterate_best = -1;
  int r,r_best = -1;
  unsigned int curr_seed = 0;
  int *cand = NULL;

  num_candidate = num_class*(num_class - 1)*(num_class - 2)/2;
  n = (max_iterate > 0) ? max_iterate : MAX_MAX_ITERATE;
  if (num_init > 1)
    n0 = (max_init_iterate > 0) ? max_init_iterate : n;
  else
    n0 = n;

  alloc_jmerge();
  alloc_jsplit();
  alloc_candidate();
  alloc_param_sm();
  alloc_ll_sm_size();
  alloc_loglike_sm();
  if (mode_daem) alloc_log_itemp_sm();
  alloc_num_iterate_sm();
  alloc_num_result_sm();
  alloc_param_part();
  alloc_prob_part();

  for (t=0; t<(num_sm+1); t++) {
    num_iterate_sm_full[t] = 0;
    num_iterate_sm_part[t] = 0;
  }

  for (t=0; t<num_sm; t++) {
    num_result_rejected[t] = 0;
    num_result_skipped[t] = 0;
  }

  for (r=0; r<num_init; r++) { /* random restarts */
    if (!mode_set_seed) {
      curr_seed = (unsigned int)random_int();
      random_seed(curr_seed);
    }

    print_msg("  [%d] #iters ",r);

    initialize_param();

    num_iterate = smem_iterate_full(0,0,n0,itemp_init);

    num_iterate_sm_full[0] = num_iterate;

    log_seed[u][r] = (!mode_set_seed) ? curr_seed : 0;
    log_num_iterate[u][r] = num_iterate;
    log_final_lambda[u][r] = final_lambda;

    if (r == 0 || final_lambda > final_lambda_best) {
      r_best = r;
      final_lambda_best = final_lambda;
      final_loglike_best = final_loglike;
      final_log_prior_best = final_log_prior;
      num_iterate_best = num_iterate;
      save_loglike_best_sm(0,num_iterate_best);
      if (mode_daem) save_log_itemp_best_sm(0,num_iterate_best);
      save_param_best();
    }
  }
  log_r_best[u] = r_best;

  final_lambda = final_lambda_best;
  final_loglike = final_loglike_best;
  final_log_prior = final_log_prior_best;
  restore_param_best();
  restore_loglike_best_sm(0,num_iterate_best);
  if (mode_daem) restore_log_itemp_best_sm(0,num_iterate_best);

  /* used for the judgement of the acceptance of a split-merge operation */
  last_lambda = final_lambda;


  /**** Split-Merge EM algorithm ****/

  for (t=0; t<num_sm; t++) {
    save_param_sm();

    compute_jmerge();
    compute_jsplit();
    sort_candidate();

    sm_ok = 0;
    num_tried_cand = 0;
    for (h=0; (max_candidate==0 || num_tried_cand<max_candidate) && h<num_candidate; h++) {

      cand = sm_candidate[h];
      k1 = cand[0];
      k2 = cand[1];
      k3 = cand[2];
      curr_jmerge = jmerge[k1][k2];
      curr_jsplit = jsplit[k3];

      if (h==0 || last_k1 != k1 || last_k2 != k2) {
        num_tried_jsplit_cand = 0;
      }
      else {
        num_tried_jsplit_cand++;
      }

      if (max_jsplit>0 && num_tried_jsplit_cand >= max_jsplit) {
#ifdef PRINT_SKIPPED
        print_msg("    -- Skipped "
                  "(# of examined split candidates reached the limit)\n");
#endif
        num_result_skipped[t]++;
        continue;
      }
      if ((min_param_c>0 && param_c[k1] < min_param_c &&
           param_c[k2] < min_param_c)
          || (min_param_c>0 && param_c[k3] < min_param_c)) {
#ifdef PRINT_SKIPPED
        print_msg("    -- Skipped (too small class probability)\n");
#endif
        num_result_skipped[t]++;
        continue;
      }

      print_msg("  [Op=%d]\n"
                "    Candidate %d (%d not skipped): "
                "Jmerge(%d,%d)=%.Nf / Jsplit(%d)=%.Nf\n",
                (t+1),h,num_tried_cand,k1,k2,curr_jmerge,k3,curr_jsplit);

      num_tried_cand++;
      last_k1 = k1;
      last_k2 = k2;

      print_msg("    #iters ");

#ifdef DISABLE_SMEM_PART
      m1 = 0;
      m2 = smem_iterate_full((t+1),0,n,1.0);
#else
      m1 = smem_iterate_part((t+1),h,0,n0);
      m2 = smem_iterate_full((t+1),m1,n0,1.0);
#endif

      /* Conditions for accepting the SM operation
       *   - log-likelihood (or log of a posteriori prob) is improved
       *   - the improvement is large enough
       *   - all class probabilities of newly obtained classes are larger than
       *     pre-defined threshold (this condition is optional)
       */
      if (final_lambda > last_lambda) {
        if ((final_lambda - last_lambda) > min_smem_improve) {
          if (!(min_new_param_c > 0) ||
              (param_c[k1] > min_new_param_c &&
               param_c[k2] > min_new_param_c &&
               param_c[k3] > min_new_param_c)) {

            print_msg("    -- Accepted [%d,%d]->[%d] & [%d]->[%d,%d]"
                      " (L current: %.Nf / last: %.Nf)\n",
                      k1,k2,k1,k3,k2,k3,final_lambda,last_lambda);
            sm_ok = 1;
          }
          else {
            print_msg("    -- Rejected [%d,%d]->[%d] & [%d]->[%d,%d]"
                      " (L current: %.Nf / last: %.Nf)\n",
                      k1,k2,k1,k3,k2,k3,final_lambda,last_lambda);
            print_msg("       <Too small split/merged class probabilities: "
                      "%d - %.Nf, %d - %.Nf, %d - %.Nf>\n",
                      k1,param_c[k1],k2,param_c[k2],k3,param_c[k3]);
            sm_ok = 0;
          }
        }
        else {
          print_msg("    -- Rejected [%d,%d]->[%d] & [%d]->[%d,%d]"
                    " (L current: %.Nf / last: %.Nf)\n",
                    k1,k2,k1,k3,k2,k3,final_lambda,last_lambda);
          print_msg("       <Too small improvement: %.Ne>\n",
                    (final_lambda - last_lambda));
          sm_ok = 0;
        }

      }
      else {
        print_msg("    -- Rejected [%d,%d]->[%d] & [%d]->[%d,%d] "
                  "(L current: %.Nf / last: %.Nf)\n",
                  k1,k2,k1,k3,k2,k3,final_lambda,last_lambda);
        print_msg("       <No improvement>\n");
        sm_ok = 0;
      }

      if (sm_ok) {
        last_lambda = final_lambda;
        num_iterate_sm_part[t+1] = m1;
        num_iterate_sm_full[t+1] = m2;
        break;
      }
      else {
        num_result_rejected[t]++;
        restore_param_sm();  /* backtrack */
      }
    }

    if (sm_ok == 0) {
      if (h == num_candidate)
        print_msg("  [Op=%d]\n  <<All candidates are rejected>>\n",(t+1));
      else
        print_msg("  [Op=%d]\n  <<The number of rejected split-merge"
                  " candidates reached the limit>>\n",(t+1));

      break;
    }
  }

  print_num_iterate_sm(fp_log,num_class);
  print_loglike_sm(fp_log,num_class);
  print_num_result_sm(fp_log,num_class);

  release_jmerge();
  release_jsplit();
  release_candidate();
  release_param_sm();
  release_loglike_sm();
  if (mode_daem) release_log_itemp_sm();
  release_ll_sm_size();
  release_num_iterate_sm();
  release_num_result_sm();
  release_param_part();
  release_prob_part();
}
